@@ -38,11 +38,6 @@ type module_type = EcAst.module_type
 
 type mod_restr = EcAst.mod_restr
 
-(*-------------------------------------------------------------------- *)
-let mhr    = EcAst.mhr
-let mleft  = EcAst.mleft
-let mright = EcAst.mright
-
 (*-------------------------------------------------------------------- *)
 let qt_equal = EcAst.qt_equal
 let qt_hash  = EcAst.qt_hash
@@ -173,13 +168,13 @@ let f_app f args ty =
 
 (* -------------------------------------------------------------------- *)
 let f_local  x ty   = mk_form (Flocal x) ty
-let f_pvar   x ty m = mk_form (Fpvar(x, m)) ty
+let f_pvar   x ty m = {m;inv=mk_form (Fpvar(x, m)) ty}
 let f_pvloc  v  m = f_pvar (pv_loc v.v_name) v.v_type m
 
 let f_pvarg  ty m = f_pvar pv_arg ty m
 
 let f_pvlocs vs menv = List.map (fun v -> f_pvloc v menv) vs
-let f_glob   m mem   = mk_form (Fglob (m, mem)) (tglob m)
+let f_glob   m mem   = {m=mem;inv=mk_form (Fglob (m, mem)) (tglob m)}
 
 (* -------------------------------------------------------------------- *)
 let f_tt     = f_op EcCoreLib.CI_Unit.p_tt    [] tunit
@@ -274,53 +269,77 @@ let f_eqs fs1 fs2 =
 let f_hoareS_r hs = mk_form (FhoareS hs) tbool
 let f_hoareF_r hf = mk_form (FhoareF hf) tbool
 
-let f_hoareS hs_m hs_pr hs_s hs_po =
-  f_hoareS_r { hs_m; hs_pr; hs_s; hs_po; }
+let f_hoareS hs_mt hs_pr hs_s hs_po =
+  assert (hs_pr.m = hs_po.m);
+  f_hoareS_r { hs_m=(hs_pr.m, hs_mt); hs_pr=hs_pr.inv; hs_s; 
+    hs_po=hs_po.inv; } [@alert "-priv_pl"]
 
-let f_hoareF hf_pr hf_f hf_po =
-  f_hoareF_r { hf_pr; hf_f; hf_po; }
+let f_hoareF pr hf_f po =
+  assert (pr.m = po.m);
+  f_hoareF_r { hf_m=pr.m; hf_pr=pr.inv; hf_f; hf_po=po.inv; } [@alert "-priv_pl"]
 
 (* -------------------------------------------------------------------- *)
 let f_eHoareS_r hs = mk_form (FeHoareS hs) tbool
 let f_eHoareF_r hf = mk_form (FeHoareF hf) tbool
 
-let f_eHoareS ehs_m ehs_pr ehs_s ehs_po =
-  f_eHoareS_r { ehs_m; ehs_pr; ehs_s; ehs_po; }
+let f_eHoareS ehs_mt ehs_pr ehs_s ehs_po =
+  assert (ehs_pr.m = ehs_po.m);
+  f_eHoareS_r { ehs_m=(ehs_pr.m, ehs_mt); ehs_pr=ehs_pr.inv; ehs_s; 
+    ehs_po=ehs_po.inv; } [@alert "-priv_pl"]
 
 let f_eHoareF ehf_pr ehf_f ehf_po =
-  f_eHoareF_r { ehf_pr; ehf_f; ehf_po; }
+  assert (ehf_pr.m = ehf_po.m);
+  f_eHoareF_r { ehf_m=ehf_pr.m; ehf_pr=ehf_pr.inv; ehf_f; ehf_po=ehf_po.inv; } [@alert "-priv_pl"]
+
+(* -------------------------------------------------------------------- *)
+
+let f_eHoare ehf_pr ehf_f ehf_po =
+  assert (ehf_pr.m = ehf_po.m);
+  f_eHoareF_r { ehf_m=ehf_pr.m; ehf_pr=ehf_pr.inv; ehf_f; ehf_po=ehf_po.inv; } [@alert "-priv_pl"]
 
 (* -------------------------------------------------------------------- *)
 let f_bdHoareS_r bhs = mk_form (FbdHoareS bhs) tbool
 let f_bdHoareF_r bhf = mk_form (FbdHoareF bhf) tbool
 
-let f_bdHoareS bhs_m bhs_pr bhs_s bhs_po bhs_cmp bhs_bd =
-  f_bdHoareS_r
-    { bhs_m; bhs_pr; bhs_s; bhs_po; bhs_cmp; bhs_bd; }
+
+let f_bdHoareS bhs_mt bhs_pr bhs_s bhs_po bhs_cmp bhs_bd =
+  assert (bhs_pr.m = bhs_po.m && bhs_bd.m = bhs_po.m);
+  f_bdHoareS_r { bhs_m=(bhs_pr.m,bhs_mt); bhs_pr=bhs_pr.inv; bhs_s; 
+    bhs_po=bhs_po.inv; bhs_cmp; bhs_bd=bhs_bd.inv; } [@alert "-priv_pl"]
 
 let f_bdHoareF bhf_pr bhf_f bhf_po bhf_cmp bhf_bd =
-  f_bdHoareF_r { bhf_pr; bhf_f; bhf_po; bhf_cmp; bhf_bd; }
+  assert (bhf_pr.m = bhf_po.m && bhf_bd.m = bhf_po.m);
+  f_bdHoareF_r { bhf_m=bhf_pr.m; bhf_pr=bhf_pr.inv; bhf_f; bhf_po=bhf_po.inv;
+                 bhf_cmp; bhf_bd=bhf_bd.inv; } [@alert "-priv_pl"]
 
 (* -------------------------------------------------------------------- *)
 let f_equivS_r es = mk_form (FequivS es) tbool
 let f_equivF_r ef = mk_form (FequivF ef) tbool
 
-let f_equivS es_ml es_mr es_pr es_sl es_sr es_po =
-   f_equivS_r { es_ml; es_mr; es_pr; es_sl; es_sr; es_po; }
+let f_equivS es_mtl es_mtr es_pr es_sl es_sr es_po =
+  assert (es_pr.ml = es_po.ml && es_pr.mr = es_po.mr);
+  let es_ml, es_mr = (es_pr.ml, es_mtl), (es_pr.mr, es_mtr) in
+  f_equivS_r { es_ml; es_mr; es_pr=es_pr.inv;
+                es_sl; es_sr; es_po=es_po.inv; } [@alert "-priv_pl"]
 
-let f_equivF ef_pr ef_fl ef_fr ef_po =
-  f_equivF_r{ ef_pr; ef_fl; ef_fr; ef_po; }
+(* -------------------------------------------------------------------- *)
+
+let f_equivF pr ef_fl ef_fr po =
+  assert (pr.ml = po.ml && pr.mr = po.mr);
+  f_equivF_r { ef_ml=pr.ml; ef_mr=pr.mr; ef_pr=pr.inv; ef_fl; ef_fr; ef_po=po.inv; } [@alert "-priv_pl"]
 
 (* -------------------------------------------------------------------- *)
 let f_eagerF_r eg = mk_form (FeagerF eg) tbool
 
 let f_eagerF eg_pr eg_sl eg_fl eg_fr eg_sr eg_po =
-  f_eagerF_r { eg_pr; eg_sl; eg_fl; eg_fr; eg_sr; eg_po; }
+  assert (eg_pr.ml = eg_po.ml && eg_pr.mr = eg_po.mr);
+  f_eagerF_r { eg_ml=eg_pr.ml; eg_mr=eg_pr.mr; eg_pr=eg_pr.inv;
+                eg_sl; eg_fl; eg_fr; eg_sr; eg_po=eg_po.inv; } [@alert "-priv_pl"]
 
 (* -------------------------------------------------------------------- *)
 let f_pr_r pr = mk_form (Fpr pr) treal
 
-let f_pr pr_mem pr_fun pr_args pr_event =
+let f_pr pr_mem pr_fun pr_args (pr_event: ss_inv) =
   f_pr_r { pr_mem; pr_fun; pr_args; pr_event; }
 
 (* -------------------------------------------------------------------- *)
@@ -432,7 +451,7 @@ let f_map gt g fp =
 
   | Fpvar (id, s) ->
       let ty' = gt fp.f_ty in
-        f_pvar id ty' s
+        (f_pvar id ty' s).inv
 
   | Fop (p, tys) ->
       let tys' = List.Smart.map gt tys in
@@ -455,56 +474,56 @@ let f_map gt g fp =
         f_proj f' i ty'
 
   | FhoareF hf ->
-      let pr' = g hf.hf_pr in
-      let po' = g hf.hf_po in
-        f_hoareF_r { hf with hf_pr = pr'; hf_po = po'; }
+      let pr' = map_ss_inv1 g (hf_pr hf) in
+      let po' = map_ss_inv1 g (hf_po hf) in
+        f_hoareF pr' hf.hf_f po'
 
   | FhoareS hs ->
-      let pr' = g hs.hs_pr in
-      let po' = g hs.hs_po in
-        f_hoareS_r { hs with hs_pr = pr'; hs_po = po'; }
+      let pr' = map_ss_inv1 g (hs_pr hs) in
+      let po' = map_ss_inv1 g (hs_po hs) in
+        f_hoareS (snd hs.hs_m) pr' hs.hs_s po'
 
   | FeHoareF hf ->
-      let pr' = g hf.ehf_pr  in
-      let po' = g hf.ehf_po  in
-      f_eHoareF_r { hf with ehf_pr = pr'; ehf_po = po' }
+      let pr' = map_ss_inv1 g (ehf_pr hf) in
+      let po' = map_ss_inv1 g (ehf_po hf) in
+        f_eHoareF pr' hf.ehf_f po'
 
   | FeHoareS hs ->
-      let pr' = g hs.ehs_pr  in
-      let po' = g hs.ehs_po  in
-        f_eHoareS_r { hs with ehs_pr = pr'; ehs_po = po'; }
+      let pr' = map_ss_inv1 g (ehs_pr hs) in
+      let po' = map_ss_inv1 g (ehs_po hs) in
+        f_eHoareS (snd hs.ehs_m) pr' hs.ehs_s po'
 
   | FbdHoareF bhf ->
-      let pr' = g bhf.bhf_pr in
-      let po' = g bhf.bhf_po in
-      let bd' = g bhf.bhf_bd in
-        f_bdHoareF_r { bhf with bhf_pr = pr'; bhf_po = po'; bhf_bd = bd'; }
+      let pr' = map_ss_inv1 g (bhf_pr bhf) in
+      let po' = map_ss_inv1 g (bhf_po bhf) in
+      let bd' = map_ss_inv1 g (bhf_bd bhf) in
+        f_bdHoareF pr' bhf.bhf_f po' bhf.bhf_cmp bd'
 
   | FbdHoareS bhs ->
-      let pr' = g bhs.bhs_pr in
-      let po' = g bhs.bhs_po in
-      let bd' = g bhs.bhs_bd in
-        f_bdHoareS_r { bhs with bhs_pr = pr'; bhs_po = po'; bhs_bd = bd'; }
+      let pr' = map_ss_inv1 g (bhs_pr bhs) in
+      let po' = map_ss_inv1 g (bhs_po bhs) in
+      let bd' = map_ss_inv1 g (bhs_bd bhs) in
+        f_bdHoareS (snd bhs.bhs_m) pr' bhs.bhs_s po' bhs.bhs_cmp bd'
 
   | FequivF ef ->
-      let pr' = g ef.ef_pr in
-      let po' = g ef.ef_po in
-        f_equivF_r { ef with ef_pr = pr'; ef_po = po'; }
+      let pr' = map_ts_inv1 g (ef_pr ef) in
+      let po' = map_ts_inv1 g (ef_po ef) in
+        f_equivF pr' ef.ef_fl ef.ef_fr po'
 
   | FequivS es ->
-      let pr' = g es.es_pr in
-      let po' = g es.es_po in
-        f_equivS_r { es with es_pr = pr'; es_po = po'; }
+      let pr' = map_ts_inv1 g (es_pr es) in
+      let po' = map_ts_inv1 g (es_po es) in
+        f_equivS (snd es.es_ml) (snd es.es_mr) pr' es.es_sl es.es_sr po'
 
   | FeagerF eg ->
-      let pr' = g eg.eg_pr in
-      let po' = g eg.eg_po in
-        f_eagerF_r { eg with eg_pr = pr'; eg_po = po'; }
+      let pr' = map_ts_inv1 g (eg_pr eg) in
+      let po' = map_ts_inv1 g (eg_po eg) in
+        f_eagerF pr' eg.eg_sl eg.eg_fl eg.eg_fr eg.eg_sr po'
 
   | Fpr pr ->
       let args' = g pr.pr_args in
-      let ev'   = g pr.pr_event in
-      f_pr_r { pr with pr_args = args'; pr_event = ev'; }
+      let ev'   = g pr.pr_event.inv in
+      f_pr_r { pr with pr_args = args'; pr_event = {m=pr.pr_event.m; inv=ev'}; }
 
 (* -------------------------------------------------------------------- *)
 let f_iter g f =
@@ -523,16 +542,16 @@ let f_iter g f =
   | Ftuple   es           -> List.iter g es
   | Fproj    (e, _)       -> g e
 
-  | FhoareF  hf   -> g hf.hf_pr; g hf.hf_po
-  | FhoareS  hs   -> g hs.hs_pr; g hs.hs_po
-  | FeHoareF  hf  -> g hf.ehf_pr; g hf.ehf_po
-  | FeHoareS  hs  -> g hs.ehs_pr; g hs.ehs_po
-  | FbdHoareF bhf -> g bhf.bhf_pr; g bhf.bhf_po; g bhf.bhf_bd
-  | FbdHoareS bhs -> g bhs.bhs_pr; g bhs.bhs_po; g bhs.bhs_bd
-  | FequivF   ef  -> g ef.ef_pr; g ef.ef_po
-  | FequivS   es  -> g es.es_pr; g es.es_po
-  | FeagerF   eg  -> g eg.eg_pr; g eg.eg_po
-  | Fpr       pr  -> g pr.pr_args; g pr.pr_event
+  | FhoareF  hf   -> g (hf_pr hf).inv; g (hf_po hf).inv
+  | FhoareS  hs   -> g (hs_pr hs).inv; g (hs_po hs).inv
+  | FeHoareF  hf  -> g (ehf_pr hf).inv; g (ehf_po hf).inv
+  | FeHoareS  hs  -> g (ehs_pr hs).inv; g (ehs_po hs).inv
+  | FbdHoareF bhf -> g (bhf_pr bhf).inv; g (bhf_po bhf).inv; g (bhf_bd bhf).inv
+  | FbdHoareS bhs -> g (bhs_pr bhs).inv; g (bhs_po bhs).inv; g (bhs_bd bhs).inv
+  | FequivF   ef  -> g (ef_pr ef).inv; g (ef_po ef).inv
+  | FequivS   es  -> g (es_pr es).inv; g (es_po es).inv
+  | FeagerF   eg  -> g (eg_pr eg).inv; g (eg_po eg).inv
+  | Fpr       pr  -> g pr.pr_args; g pr.pr_event.inv
 
 
 (* -------------------------------------------------------------------- *)
@@ -552,16 +571,16 @@ let form_exists g f =
   | Ftuple   es           -> List.exists g es
   | Fproj    (e, _)       -> g e
 
-  | FhoareF   hf -> g hf.hf_pr   || g hf.hf_po
-  | FhoareS   hs -> g hs.hs_pr   || g hs.hs_po
-  | FeHoareF  hf  -> g hf.ehf_pr || g hf.ehf_po
-  | FeHoareS  hs  -> g hs.ehs_pr || g hs.ehs_po
-  | FbdHoareF bhf -> g bhf.bhf_pr  || g bhf.bhf_po
-  | FbdHoareS bhs -> g bhs.bhs_pr  || g bhs.bhs_po
-  | FequivF   ef  -> g ef.ef_pr    || g ef.ef_po
-  | FequivS   es  -> g es.es_pr    || g es.es_po
-  | FeagerF   eg  -> g eg.eg_pr    || g eg.eg_po
-  | Fpr       pr  -> g pr.pr_args  || g pr.pr_event
+  | FhoareF   hf -> g (hf_pr hf).inv   || g (hf_po hf).inv
+  | FhoareS   hs -> g (hs_pr hs).inv   || g (hs_po hs).inv
+  | FeHoareF  hf  -> g (ehf_pr hf).inv || g (ehf_po hf).inv
+  | FeHoareS  hs  -> g (ehs_pr hs).inv || g (ehs_po hs).inv
+  | FbdHoareF bhf -> g (bhf_pr bhf).inv || g (bhf_po bhf).inv
+  | FbdHoareS bhs -> g (bhs_pr bhs).inv || g (bhs_po bhs).inv
+  | FequivF   ef  -> g (ef_pr ef).inv   || g (ef_po ef).inv
+  | FequivS   es  -> g (es_pr es).inv   || g (es_po es).inv
+  | FeagerF   eg  -> g (eg_pr eg).inv    || g (eg_po eg).inv
+  | Fpr       pr  -> g pr.pr_args  || g pr.pr_event.inv
 
 (* -------------------------------------------------------------------- *)
 let form_forall g f =
@@ -580,16 +599,16 @@ let form_forall g f =
   | Ftuple   es           -> List.for_all g es
   | Fproj    (e, _)       -> g e
 
-  | FhoareF  hf  -> g hf.hf_pr  && g hf.hf_po
-  | FhoareS  hs  -> g hs.hs_pr  && g hs.hs_po
-  | FbdHoareF bhf -> g bhf.bhf_pr && g bhf.bhf_po
-  | FbdHoareS bhs -> g bhs.bhs_pr && g bhs.bhs_po
-  | FequivF   ef  -> g ef.ef_pr   && g ef.ef_po
-  | FequivS   es  -> g es.es_pr   && g es.es_po
-  | FeagerF   eg  -> g eg.eg_pr   && g eg.eg_po
-  | Fpr       pr  -> g pr.pr_args && g pr.pr_event
-  | FeHoareF  hf  -> g hf.ehf_pr && g hf.ehf_po
-  | FeHoareS  hs  -> g hs.ehs_pr && g hs.ehs_po
+  | FhoareF  hf  -> g (hf_pr hf).inv  && g (hf_po hf).inv
+  | FhoareS  hs  -> g (hs_pr hs).inv  && g (hs_po hs).inv
+  | FbdHoareF bhf -> g (bhf_pr bhf).inv && g (bhf_po bhf).inv
+  | FbdHoareS bhs -> g (bhs_pr bhs).inv && g (bhs_po bhs).inv
+  | FequivF   ef  -> g (ef_pr ef).inv   && g (ef_po ef).inv
+  | FequivS   es  -> g (es_pr es).inv   && g (es_po es).inv
+  | FeagerF   eg  -> g (eg_pr eg).inv   && g (eg_po eg).inv
+  | Fpr       pr  -> g pr.pr_args && g pr.pr_event.inv
+  | FeHoareF  hf  -> g (ehf_pr hf).inv && g (ehf_po hf).inv
+  | FeHoareS  hs  -> g (ehs_pr hs).inv && g (ehs_po hs).inv
 
 
 (* -------------------------------------------------------------------- *)
@@ -809,6 +828,16 @@ let destr_imp = destr_app2 ~name:"imp" is_op_imp
 let destr_iff = destr_app2 ~name:"iff" is_op_iff
 let destr_eq  = destr_app2 ~name:"eq"  is_op_eq
 
+let destr_and_ts_inv inv = 
+  let c1 = map_ts_inv1 (fun po -> fst (destr_and po)) inv in
+  let c2 = map_ts_inv1 (fun po -> snd (destr_and po)) inv in
+  (c1, c2)
+
+let destr_and_ss_inv inv =
+  let c1 = map_ss_inv1 (fun po -> fst (destr_and po)) inv in
+  let c2 = map_ss_inv1 (fun po -> snd (destr_and po)) inv in
+  (c1, c2)
+
 let destr_and3 f =
   try
     let c1, (c2, c3) = snd_map destr_and (destr_and f)
@@ -898,7 +927,8 @@ let equantif_of_quantif (qt : quantif) : equantif =
   | Lexists -> `EExists
 
 (* -------------------------------------------------------------------- *)
-let rec form_of_expr mem (e : expr) =
+
+let rec form_of_expr_r ?m (e : expr) =
   match e.e_node with
   | Eint n ->
      f_int n
@@ -907,44 +937,53 @@ let rec form_of_expr mem (e : expr) =
      f_local id e.e_ty
 
   | Evar pv ->
-     f_pvar pv e.e_ty mem
+    begin
+     match m with
+     | None -> failwith "expecting memory"
+     | Some m -> (f_pvar pv e.e_ty m).inv
+    end
 
   | Eop (op, tys) ->
      f_op op tys e.e_ty
 
   | Eapp (ef, es) ->
-     f_app (form_of_expr mem ef) (List.map (form_of_expr mem) es) e.e_ty
+     f_app (form_of_expr_r ?m ef) (List.map (form_of_expr_r ?m) es) e.e_ty
 
   | Elet (lpt, e1, e2) ->
-     f_let lpt (form_of_expr mem e1) (form_of_expr mem e2)
+     f_let lpt (form_of_expr_r ?m e1) (form_of_expr_r ?m e2)
 
   | Etuple es ->
-     f_tuple (List.map (form_of_expr mem) es)
+     f_tuple (List.map (form_of_expr_r ?m) es)
 
   | Eproj (e1, i) ->
-     f_proj (form_of_expr mem e1) i e.e_ty
+     f_proj (form_of_expr_r ?m e1) i e.e_ty
 
   | Eif (e1, e2, e3) ->
-     let e1 = form_of_expr mem e1 in
-     let e2 = form_of_expr mem e2 in
-     let e3 = form_of_expr mem e3 in
+     let e1 = form_of_expr_r ?m e1 in
+     let e2 = form_of_expr_r ?m e2 in
+     let e3 = form_of_expr_r ?m e3 in
      f_if e1 e2 e3
 
   | Ematch (b, fs, ty) ->
-     let b'  = form_of_expr mem b in
-     let fs' = List.map (form_of_expr mem) fs in
+     let b'  = form_of_expr_r ?m b in
+     let fs' = List.map (form_of_expr_r ?m) fs in
      f_match b' fs' ty
 
   | Equant (qt, b, e) ->
      let b = List.map (fun (x, ty) -> (x, GTty ty)) b in
-     let e = form_of_expr mem e in
+     let e = form_of_expr_r ?m e in
      f_quant (quantif_of_equantif qt) b e
 
+let form_of_expr e = form_of_expr_r e
+
+let ss_inv_of_expr m (e : expr) =
+  {m;inv=form_of_expr_r ~m e}
 
 (* -------------------------------------------------------------------- *)
 exception CannotTranslate
 
-let expr_of_form mh f =
+let expr_of_ss_inv f =
+  let mh, f = f.m, f.inv in
   let rec aux fp =
     match fp.f_node with
     | Fint   z -> e_int z
@@ -986,6 +1025,43 @@ let expr_of_form mh f =
 
   in aux f
 
+let expr_of_form f =
+  let rec aux fp =
+    match fp.f_node with
+    | Fint   z -> e_int z
+    | Flocal x -> e_local x fp.f_ty
+
+    | Fop  (p, tys) -> e_op p tys fp.f_ty
+    | Fapp (f, fs)  -> e_app (aux f) (List.map aux fs) fp.f_ty
+    | Ftuple fs     -> e_tuple (List.map aux fs)
+    | Fproj  (f, i) -> e_proj (aux f) i fp.f_ty
+
+    | Fif (c, f1, f2) ->
+      e_if (aux c) (aux f1) (aux f2)
+
+    | Fmatch (c, bs, ty) ->
+      e_match (aux c) (List.map aux bs) ty
+
+    | Flet (lp, f1, f2) ->
+      e_let lp (aux f1) (aux f2)
+
+    | Fquant (kd, bds, f) ->
+      e_quantif (equantif_of_quantif kd) (List.map auxbd bds) (aux f)
+
+    | Fpvar     _ | Fglob     _
+    | FhoareF   _ | FhoareS   _
+    | FeHoareF  _ | FeHoareS  _
+    | FbdHoareF _ | FbdHoareS _
+    | FequivF   _ | FequivS   _
+    | FeagerF   _ | Fpr       _ -> raise CannotTranslate
+
+  and auxbd ((x, bd) : binding) =
+    match bd with
+    | GTty ty -> (x, ty)
+    | _ -> raise CannotTranslate
+
+  in aux f
+
 (* -------------------------------------------------------------------- *)
 (* A predicate on memory: λ mem. -> pred *)
 type mem_pr = EcMemory.memory * form
