@@ -64,16 +64,20 @@ type meerror =
 exception MEError of meerror
 
 module Memory : sig
-  val all         : env -> memenv list
-  val set_active  : memory -> env -> env
-  val get_active  : env -> memory option
+  val all           : env -> memenv list
+  val set_active_ss : memory -> env -> env
+  val get_active_ss : env -> memory option
+  val set_active_ts : memory -> memory -> env -> env
+  val get_active_ts : env -> (memory * memory) option
 
-  val byid        : memory -> env -> memenv option
-  val lookup      : symbol -> env -> memenv option
-  val current     : env -> memenv option
-  val push        : memenv -> env -> env
-  val push_all    : memenv list -> env -> env
-  val push_active : memenv -> env -> env
+  val byid          : memory -> env -> memenv option
+  val lookup        : symbol -> env -> memenv option
+  val current_ss    : env -> memenv option
+  val current_ts    : env -> (memenv * memenv) option
+  val push          : memenv -> env -> env
+  val push_all      : memenv list -> env -> env
+  val push_active_ss: memenv -> env -> env
+  val push_active_ts: memenv -> memenv -> env -> env
 end
 
 (* -------------------------------------------------------------------- *)
@@ -93,29 +97,27 @@ module Fun : sig
   val add   : xpath -> env -> env
 
   (* ------------------------------------------------------------------ *)
-  val prF_memenv : EcMemory.memory -> xpath -> env -> memenv
+  val prF_memenv : memory -> xpath -> env -> memenv
 
-  val prF : xpath -> env -> env
+  val prF : memory -> xpath -> env -> env
 
-  val hoareF_memenv : xpath -> env -> memenv * memenv
+  val hoareF_memenv : memory -> xpath -> env -> memenv * memenv
 
-  val hoareF : xpath -> env -> env * env
+  val hoareF : memory -> xpath -> env -> env * env
 
-  val hoareS : xpath -> env -> memenv * (funsig * function_def) * env
+  val hoareS : memory -> xpath -> env -> memenv * (funsig * function_def) * env
 
   val actmem_body :  memory -> function_ -> (funsig * function_def) * memenv
   val actmem_post :  memory -> function_ -> memenv
 
-  val inv_memory : [`Left|`Right] -> memenv
+  val inv_memenv : memory -> memory -> env -> env
 
-  val inv_memenv : env -> env
-
-  val equivF_memenv : xpath -> xpath -> env ->
+  val equivF_memenv : memory -> memory -> xpath -> xpath -> env ->
     (memenv * memenv) * (memenv * memenv)
 
-  val equivF : xpath -> xpath -> env -> env * env
+  val equivF : memory -> memory -> xpath -> xpath -> env -> env * env
 
-  val equivS : xpath -> xpath -> env ->
+  val equivS : memory -> memory -> xpath -> xpath -> env ->
     memenv * (funsig * function_def) * memenv * (funsig * function_def) * env
 end
 
@@ -246,7 +248,7 @@ module NormMp : sig
 
   val flatten_use : use -> EcIdent.t list * (xpath * ty) list
 
-  val norm_glob     : env -> EcMemory.memory -> mpath -> form
+  val norm_glob     : env -> EcMemory.memory -> mpath -> ss_inv
   val norm_tglob    : env -> mpath -> EcTypes.ty
 
   val is_abstract_fun : xpath -> env -> bool
@@ -502,14 +504,15 @@ module LDecl : sig
 
   val clear : ?leniant:bool -> EcIdent.Sid.t -> hyps -> hyps
 
-  val push_all    : memenv list -> hyps -> hyps
-  val push_active : memenv -> hyps -> hyps
+  val push_all       : memenv list -> hyps -> hyps
+  val push_active_ss : memenv -> hyps -> hyps
+  val push_active_ts : memenv -> memenv -> hyps -> hyps
 
-  val hoareF : xpath -> hyps -> hyps * hyps
-  val equivF : xpath -> xpath -> hyps -> hyps * hyps
+  val hoareF : memory -> xpath -> hyps -> hyps * hyps
+  val equivF : memory -> memory -> xpath -> xpath -> hyps -> hyps * hyps
 
-  val inv_memenv  : hyps -> hyps
-  val inv_memenv1 : hyps -> hyps
+  val inv_memenv  : memory -> memory -> hyps -> hyps
+  val inv_memenv1 : memory -> hyps -> hyps
 end
 
-val pp_debug_form : (env -> Format.formatter -> form -> unit) ref
+val pp_debug_form : (env -> form -> unit) ref
