@@ -202,12 +202,12 @@ and pformula_r =
   | PFlsless  of pgamepath
   | PFscope   of pqsymbol * pformula
 
-  | PFhoareF   of pformula * pgamepath * pformula
-  | PFehoareF  of pformula * pgamepath * pformula
-  | PFequivF   of pformula * (pgamepath * pgamepath) * pformula
-  | PFeagerF   of pformula * (pstmt * pgamepath * pgamepath * pstmt) * pformula
-  | PFprob     of pgamepath * (pformula list) * pmemory * pformula
-  | PFBDhoareF of pformula * pgamepath * pformula * phoarecmp * pformula
+  | PFhoareF   of psymbol option * pformula * pgamepath * pformula
+  | PFehoareF  of psymbol option * pformula * pgamepath * pformula
+  | PFequivF   of psymbol option * psymbol option * pformula * (pgamepath * pgamepath) * pformula
+  | PFeagerF   of psymbol option * psymbol option * pformula * (pstmt * pgamepath * pgamepath * pstmt) * pformula
+  | PFprob     of psymbol option * pgamepath * (pformula list) * pmemory * pformula
+  | PFBDhoareF of psymbol option * pformula * pgamepath * pformula * phoarecmp * pformula
 
 and pmemtype_el = ([`Single|`Tuple] * (psymbol list)) located * pty
 and pmemtype    = pmemtype_el list
@@ -770,8 +770,10 @@ type phltactic =
   | Prw_equiv      of rw_eqv_info
   | Psymmetry
   | Pbdhoare_split of bdh_split
-  | Pprocchange    of side option * pcodepos * pexpr
+  | Prwprgm of rwprgm
   | Pprocrewrite   of side option * pcodepos * prrewrite
+  | Pchangestmt    of side option * pcodepos_range * pstmt
+
 
     (* Eager *)
   | Peager_seq       of (eager_info * pcodepos1 pair * pformula)
@@ -789,6 +791,10 @@ type phltactic =
   | Pauto
   | Plossless
 
+and rwprgm = [
+  | `IdAssign of pcodepos * pqsymbol
+]
+
 (* -------------------------------------------------------------------- *)
 type include_exclude = [ `Include | `Exclude ]
 type pdbmap1 = {
