@@ -1143,6 +1143,7 @@ module Op = struct
       | `Plain e  -> Some (OP_Plain (fs e))
       | `Fix opfx ->
           Some (OP_Fix {
+            opf_recp     = EcPath.pqname (EcEnv.root eenv) (EcIdent.name opfx.EHI.mf_name);
             opf_args     = opfx.EHI.mf_args;
             opf_resty    = opfx.EHI.mf_codom;
             opf_struct   = (opfx.EHI.mf_recs, List.length opfx.EHI.mf_args);
@@ -1345,7 +1346,7 @@ module Op = struct
       (`Det, Sem.translate_e env ret) in
 
     let mode, aout = Sem.translate_s env cont body.f_body in
-    let aout = form_of_expr mhr aout in (* FIXME: translate to forms directly? *)
+    let aout = form_of_expr aout in (* FIXME: translate to forms directly? *)
     let aout = f_lambda (List.map2 (fun (_, ty) x -> (x, GTty ty)) params ids) aout in
 
     let opdecl = EcDecl.{
@@ -1365,10 +1366,10 @@ module Op = struct
     let scope =
       let prax =
         let locs  = List.map (fun (x, ty) -> (EcIdent.create x, ty)) params in
-        let res   = f_pvar pv_res sig_.fs_ret mhr in
+        let prmem = EcIdent.create "&m" in
+        let res   = f_pvar pv_res sig_.fs_ret prmem in
         let resx  = EcIdent.create "v" in
         let resv  = f_local resx sig_.fs_ret in
-        let prmem = EcIdent.create "&m" in
 
         let mu =
           let sem =
@@ -1393,7 +1394,7 @@ module Op = struct
                 (f_pr prmem
                    f
                    (f_tuple (List.map (fun (x, ty) -> f_local x ty) locs))
-                   (f_eq res resv))
+                   (map_ss_inv1 (fun r -> f_eq r resv) res))
                 mu))
       in
 
@@ -1412,22 +1413,23 @@ module Op = struct
       | `Det ->
          let hax =
            let locs  = List.map (fun (x, ty) -> (EcIdent.create x, ty)) params in
-           let res   = f_pvar pv_res sig_.fs_ret mhr in
-           let args  = f_pvar pv_arg sig_.fs_arg mhr in
+           let m = EcIdent.create "&hr" in
+           let res   = f_pvar pv_res sig_.fs_ret m in
+           let args  = f_pvar pv_arg sig_.fs_arg m in
 
            f_forall
              (List.map (fun (x, ty) -> (x, GTty ty)) locs)
              (f_hoareF
-                (f_eq
-                   args
-                   (f_tuple (List.map (fun (x, ty) -> f_local x ty) locs)))
+                {m;inv=(f_eq
+                   args.inv
+                   (f_tuple (List.map (fun (x, ty) -> f_local x ty) locs)))}
                 f
-                (f_eq
-                   res
+                {m;inv=(f_eq
+                   res.inv
                    (f_app
                       (f_op oppath [] opdecl.op_ty)
                       (List.map (fun (x, ty) -> f_local x ty) locs)
-                      sig_.fs_ret)))
+                      sig_.fs_ret))})
          in
 
          let prax = EcDecl.{
@@ -2419,7 +2421,7 @@ end
                     let tip = f_subst_init ~tv:tip () in
                     let es = e_subst tip in
                     let xs  = List.map (snd_map (ty_subst tip)) nt.ont_args in
-                    let bd  = EcFol.form_of_expr EcFol.mhr (es nt.ont_body) in
+                    let bd  = EcFol.form_of_expr (es nt.ont_body) in
                     let fp  = EcFol.f_lambda (List.map (snd_map EcFol.gtty) xs) bd in
 
                     match fp.f_node with
