@@ -50,200 +50,271 @@ let keep_of_mode (mode : clmode) =
 (* -------------------------------------------------------------------- *)
 exception Incompatible of incompatible
 
-let tparams_compatible rtyvars ntyvars =
-  let rlen = List.length rtyvars and nlen = List.length ntyvars in
-  if rlen <> nlen then
-    raise (Incompatible (NotSameNumberOfTyParam(rlen,nlen)))
-
-let ty_compatible env ue (rtyvars, rty) (ntyvars, nty) =
-  tparams_compatible rtyvars ntyvars;
-  let subst = CS.Tvar.init rtyvars (List.map tvar ntyvars) in
-  let rty   = CS.Tvar.subst subst rty in
-  try  EcUnify.unify env ue rty nty
-  with EcUnify.UnificationFailure _ ->
-    raise (Incompatible (DifferentType (rty, nty)))
-
 (* -------------------------------------------------------------------- *)
 let error_body exn b = if not b then raise exn
 
 (* -------------------------------------------------------------------- *)
-let ri_compatible =
-    { EcReduction.full_red with delta_p = (fun _-> `Force); user = false }
-
-(* -------------------------------------------------------------------- *)
-let constr_compatible exn env cs1 cs2 =
-  error_body exn (List.length cs1 = List.length cs2);
-  let doit (s1,tys1) (s2,tys2) =
-    error_body exn (EcSymbols.sym_equal s1 s2);
-    error_body exn (List.length tys1 = List.length tys2);
-    List.iter2 (fun ty1 ty2 -> error_body exn (EcReduction.EqTest.for_type env ty1 ty2)) tys1 tys2 in
-  List.iter2 doit cs1 cs2
-
-let datatype_compatible exn hyps ty1 ty2 =
-  let env = EcEnv.LDecl.toenv hyps in
-  constr_compatible exn env ty1.tydt_ctors ty2.tydt_ctors
-
-let record_compatible exn hyps f1 pr1 f2 pr2 =
-  error_body exn (EcReduction.is_conv hyps f1 f2);
-  error_body exn (List.length pr1 = List.length pr2);
-  let env = EcEnv.LDecl.toenv hyps in
-  let doit (s1,ty1) (s2,ty2) =
-    error_body exn (EcSymbols.sym_equal s1 s2);
-    error_body exn (EcReduction.EqTest.for_type env ty1 ty2) in
-  List.iter2 doit pr1 pr2
-
-let get_open_tydecl hyps p tys =
-  let tydecl = EcEnv.Ty.by_path p (EcEnv.LDecl.toenv hyps) in
+let get_open_tydecl (env : EcEnv.env) (p : EcPath.path) (tys : ty list) =
+  let tydecl = EcEnv.Ty.by_path p env in
   EcSubst.open_tydecl tydecl tys
 
-let rec tybody_compatible exn hyps ty_body1 ty_body2 =
-  match ty_body1, ty_body2 with
-  | `Abstract _, `Abstract _ -> () (* FIXME Sp.t *)
-  | `Concrete ty1   , `Concrete ty2 -> error_body exn (EcReduction.EqTest.for_type (EcEnv.LDecl.toenv hyps) ty1 ty2)
-  | `Datatype ty1   , `Datatype ty2 -> datatype_compatible exn hyps ty1 ty2
-  | `Record (f1,pr1), `Record(f2,pr2) -> record_compatible exn hyps f1 pr1 f2 pr2
-  | _, `Concrete {ty_node = Tconstr(p, tys) } ->
-    let ty_body2 = get_open_tydecl hyps p tys in
-    tybody_compatible exn hyps ty_body1 ty_body2
-  | `Concrete{ty_node = Tconstr(p, tys) }, _ ->
-    let ty_body1 = get_open_tydecl hyps p tys in
-    tybody_compatible exn hyps ty_body1 ty_body2
-  | _, _ -> raise exn (* FIXME should we do more for concrete version other *)
-
-let tydecl_compatible env tyd1 tyd2 =
-  let params = tyd1.tyd_params in
-  tparams_compatible params tyd2.tyd_params;
-  let tparams = List.map (fun (id,_) -> tvar id) params in
-  let ty_body1 = tyd1.tyd_type in
-  let ty_body2 = EcSubst.open_tydecl tyd2 tparams in
-  let exn  = Incompatible (TyBody(*tyd1,tyd2*)) in
-  let hyps = EcEnv.LDecl.init env params in
-  match ty_body1, ty_body2 with
-  | `Abstract _, _ -> () (* FIXME Sp.t *)
-  | _, _ -> tybody_compatible exn hyps ty_body1 ty_body2
+(* -------------------------------------------------------------------- *)
+exception CoreIncompatible
 
 (* -------------------------------------------------------------------- *)
-let expr_compatible exn env s e1 e2 =
-  let f1 = EcFol.form_of_expr EcFol.mhr e1 in
-  let f2 = EcSubst.subst_form s (EcFol.form_of_expr EcFol.mhr e2) in
-  error_body exn (EcReduction.is_conv ~ri:ri_compatible (EcEnv.LDecl.init env []) f1 f2)
-
-let get_open_oper exn env p tys =
+let get_open_oper (env : EcEnv.env) (p : EcPath.path) (tys : ty list) =
   let oper = EcEnv.Op.by_path p env in
   let _, okind = EcSubst.open_oper oper tys in
   match okind with
   | OB_oper (Some ob) -> ob
-  | _ -> raise exn
+  | _ -> raise CoreIncompatible
 
-let rec oper_compatible exn env ob1 ob2 =
-  match ob1, ob2 with
-  | OP_Plain f1, OP_Plain f2 ->
-    error_body exn (EcReduction.is_conv ~ri:ri_compatible (EcEnv.LDecl.init env []) f1 f2)
-  | OP_Plain {f_node = Fop(p,tys)}, _ ->
-    let ob1 = get_open_oper exn env p tys  in
-    oper_compatible exn env ob1 ob2
-  | _, OP_Plain {f_node = Fop(p,tys)} ->
-    let ob2 = get_open_oper exn env p tys in
-    oper_compatible exn env ob1 ob2
-  | OP_Constr(p1,i1), OP_Constr(p2,i2) ->
-    error_body exn (EcPath.p_equal p1 p2 && i1 = i2)
-  | OP_Record p1, OP_Record p2 ->
-    error_body exn (EcPath.p_equal p1 p2)
-  | OP_Proj(p1,i11,i12), OP_Proj(p2,i21,i22) ->
-    error_body exn (EcPath.p_equal p1 p2 && i11 = i21 && i12 = i22)
-  | OP_Fix f1, OP_Fix f2 ->
-    opfix_compatible exn env f1 f2
-  | OP_TC, OP_TC -> ()
-  | _, _ -> raise exn
-
-and opfix_compatible exn env f1 f2 =
-  let s = params_compatible exn env EcSubst.empty f1.opf_args f2.opf_args in
-  error_body exn (EcReduction.EqTest.for_type env f1.opf_resty f2.opf_resty);
-  error_body exn (f1.opf_struct = f2.opf_struct);
-  opbranches_compatible exn env s f1.opf_branches f2.opf_branches
-
-and params_compatible exn env s p1 p2 =
-  error_body exn (List.length p1 = List.length p2);
-  let doit s (id1,ty1) (id2,ty2) =
-    error_body exn (EcReduction.EqTest.for_type env ty1 ty2);
-    EcSubst.add_flocal s id2 (EcFol.f_local id1 ty1) in
-  List.fold_left2 doit s p1 p2
-
-and opbranches_compatible exn env s ob1 ob2 =
-  match ob1, ob2 with
-  | OPB_Leaf(d1,e1), OPB_Leaf(d2,e2) ->
-    error_body exn (List.length d1 = List.length d2);
-    let s =
-      List.fold_left2 (params_compatible exn env) s d1 d2 in
-    expr_compatible exn env s e1 e2
-
-  | OPB_Branch obs1, OPB_Branch obs2 ->
-    error_body exn (Parray.length obs1 = Parray.length obs2);
-    Parray.iter2 (opbranch_compatible exn env s) obs1 obs2
-  | _, _ -> raise exn
-
-and opbranch_compatible exn env s ob1 ob2 =
-  error_body exn (EcPath.p_equal (fst ob1.opb_ctor) (fst ob2.opb_ctor));
-  error_body exn (snd ob1.opb_ctor = snd ob2.opb_ctor);
-  opbranches_compatible exn env s ob1.opb_sub ob2.opb_sub
-
-let get_open_pred exn env p tys =
+(* -------------------------------------------------------------------- *)
+let get_open_pred (env : EcEnv.env) (p : EcPath.path) (tys : ty list) =
   let oper = EcEnv.Op.by_path p env in
   let _, okind = EcSubst.open_oper oper tys in
   match okind with
   | OB_pred (Some pb) -> pb
-  | _ -> raise exn
+  | _ -> raise CoreIncompatible
 
-let rec pred_compatible exn env pb1 pb2 =
-  match pb1, pb2 with
-  | PR_Plain f1, PR_Plain f2 -> error_body exn (EcReduction.is_conv (EcEnv.LDecl.init env []) f1 f2)
-  | PR_Plain {f_node = Fop(p,tys)}, _ ->
-    let pb1 = get_open_pred exn env p tys  in
-    pred_compatible exn env pb1 pb2
-  | _, PR_Plain {f_node = Fop(p,tys)} ->
-    let pb2 = get_open_pred exn env p tys  in
-    pred_compatible exn env pb1 pb2
-  | PR_Ind pr1, PR_Ind pr2 ->
-    ind_compatible exn env pr1 pr2
-  | _, _ -> raise exn
+(* -------------------------------------------------------------------- *)
+module Compatible : sig
+  type 'a comparator = EcEnv.env -> 'a -> 'a -> unit
 
-and ind_compatible exn env pi1 pi2 =
-  let s = params_compatible exn env EcSubst.empty pi1.pri_args pi2.pri_args in
-  error_body exn (List.length pi1.pri_ctors = List.length pi2.pri_ctors);
-  List.iter2 (prctor_compatible exn env s) pi1.pri_ctors pi2.pri_ctors
+  val for_ty :
+       EcEnv.env
+    -> EcUnify.unienv
+    -> EcIdent.ident list * ty
+    -> EcIdent.ident list * ty
+    -> unit
 
-and prctor_compatible exn env s prc1 prc2 =
-  error_body exn (EcSymbols.sym_equal prc1.prc_ctor prc2.prc_ctor);
-  let env, s = EcReduction.check_bindings exn env s prc1.prc_bds prc2.prc_bds in
-  error_body exn (List.length prc1.prc_spec = List.length prc2.prc_spec);
-  let doit f1 f2 =
-    error_body exn (EcReduction.is_conv (EcEnv.LDecl.init env []) f1 (EcSubst.subst_form s f2)) in
-  List.iter2 doit prc1.prc_spec prc2.prc_spec
+  val for_tydecl   : tydecl comparator
+  val for_operator : operator comparator
+end = struct
+  open EcEnv.LDecl
 
-let nott_compatible exn env nb1 nb2 =
-  let s = params_compatible exn env EcSubst.empty nb1.ont_args nb2.ont_args in
-  (* We do not check ont_resty because it is redundant *)
-  expr_compatible exn env s nb1.ont_body nb2.ont_body
+  type 'a comparator = EcEnv.env -> 'a -> 'a -> unit
 
-let operator_compatible env oper1 oper2 =
-  let open EcDecl in
-  let params = oper1.op_tparams in
-  tparams_compatible oper1.op_tparams oper2.op_tparams;
-  let oty1, okind1 = oper1.op_ty, oper1.op_kind in
-  let tparams = List.map (fun (id,_) -> tvar id) params in
-  let oty2, okind2 = EcSubst.open_oper oper2 tparams in
-  if not (EcReduction.EqTest.for_type env oty1 oty2) then
-    raise (Incompatible (DifferentType(oty1, oty2)));
-  let hyps = EcEnv.LDecl.init env params in
-  let env  = EcEnv.LDecl.toenv hyps in
-  let exn  = Incompatible (OpBody(*oper1,oper2*)) in
-  match okind1, okind2 with
-  | OB_oper None      , OB_oper _          -> ()
-  | OB_oper (Some ob1), OB_oper (Some ob2) -> oper_compatible exn env ob2 ob1
-  | OB_pred None      , OB_pred _          -> ()
-  | OB_pred (Some pb1), OB_pred (Some pb2) -> pred_compatible exn env pb2 pb1
-  | OB_nott nb1       , OB_nott nb2        -> nott_compatible exn env nb2 nb1
-  | _                 , _                  -> raise exn
+  let ri_compatible =
+    { EcReduction.full_red with delta_p = (fun _-> `Force); user = false }
+
+  let check (b : bool) =
+    if not b then raise CoreIncompatible
+
+  let for_tparams rtyvars ntyvars =
+    let rlen = List.length rtyvars
+    and nlen = List.length ntyvars in
+
+    if rlen <> nlen then
+      raise (Incompatible (NotSameNumberOfTyParam (rlen, nlen)))
+
+  let for_params 
+    (hyps : hyps)
+    (s    : EcSubst.subst)
+    (p1   : (EcIdent.ident * ty) list)
+    (p2   : (EcIdent.ident * ty) list)
+  =
+    check (List.compare_lengths p1 p2 = 0);
+
+    let do_param s (id1, ty1) (id2, ty2) =
+      check (EcReduction.EqTest.for_type (toenv hyps) ty1 ty2);
+      EcSubst.add_flocal s id2 (EcFol.f_local id1 ty1)
+    in List.fold_left2 do_param s p1 p2
+
+  let for_ty (env : EcEnv.env) (ue : EcUnify.unienv) (rtyvars, rty) (ntyvars, nty) =
+    for_tparams rtyvars ntyvars;
+
+    let subst = CS.Tvar.init rtyvars (List.map tvar ntyvars) in
+    let rty   = CS.Tvar.subst subst rty in
+
+    try  EcUnify.unify env ue rty nty
+    with EcUnify.UnificationFailure _ ->
+      raise (Incompatible (DifferentType (rty, nty)))
+
+  let for_expr (hyps : hyps) (s : EcSubst.subst) (e1 : expr) (e2 : expr) =
+    let f1 = EcFol.form_of_expr e1 in
+    let f2 = EcSubst.subst_form s (EcFol.form_of_expr e2) in
+    check (EcReduction.is_conv ~ri:ri_compatible hyps f1 f2)
+
+  let for_datatype (hyps : hyps) (ty1 : ty_dtype) (ty2 : ty_dtype) =
+    let for_constr (cs1 : ty_dtype_ctor list) (cs2 : ty_dtype_ctor list) =
+      check (List.compare_lengths cs1 cs2 = 0);
+
+      let for_ctor1 (s1,tys1) (s2,tys2) =
+        check (EcSymbols.sym_equal s1 s2);
+        check (List.compare_lengths tys1 tys2 = 0);
+        List.iter2 (fun ty1 ty2 ->
+          check (EcReduction.EqTest.for_type (toenv hyps) ty1 ty2)
+        ) tys1 tys2
+      in List.iter2 for_ctor1 cs1 cs2
+    in for_constr ty1.tydt_ctors ty2.tydt_ctors
+
+  let for_record (hyps : hyps) ((f1, pr1) : ty_record) ((f2, pr2) : ty_record) =
+    check (EcReduction.is_conv hyps f1 f2);
+
+    let for_field (s1, ty1) (s2, ty2) =
+      check (EcSymbols.sym_equal s1 s2);
+      check (EcReduction.EqTest.for_type (toenv hyps) ty1 ty2)
+    in List.iter2 for_field pr1 pr2
+
+  let rec tybody (hyps : EcEnv.LDecl.hyps) (ty_body1 : ty_body) (ty_body2 : ty_body) =
+    match ty_body1, ty_body2 with
+    | `Abstract _   ,  `Abstract _   -> () (* FIXME Sp.t *)
+    | `Concrete ty1 , `Concrete ty2  -> check (EcReduction.EqTest.for_type (toenv hyps) ty1 ty2)
+    | `Datatype ty1 , `Datatype ty2  -> for_datatype hyps ty1 ty2
+    | `Record   rec1, `Record   rec2 -> for_record hyps rec1 rec2
+
+    | _, `Concrete { ty_node = Tconstr (p, tys) } ->
+      let ty_body2 = get_open_tydecl (toenv hyps) p tys in
+      tybody hyps ty_body1 ty_body2
+
+    | `Concrete{ ty_node = Tconstr (p, tys) }, _ ->
+      let ty_body1 = get_open_tydecl (toenv hyps) p tys in
+      tybody hyps ty_body1 ty_body2
+
+    | _, _ -> raise CoreIncompatible
+
+  let for_tydecl (env : EcEnv.env) (tyd1 : tydecl) (tyd2 : tydecl) =
+    try
+      let params = tyd1.tyd_params in
+
+      for_tparams params tyd2.tyd_params;
+
+      let tparams = List.map (fun (id,_) -> tvar id) params in
+      let ty_body1 = tyd1.tyd_type in
+      let ty_body2 = EcSubst.open_tydecl tyd2 tparams in
+
+      let hyps = EcEnv.LDecl.init env params in
+
+      match ty_body1, ty_body2 with
+      | `Abstract _, _ -> () (* FIXME Sp.t *)
+      | _, _ -> tybody hyps ty_body1 ty_body2
+
+    with CoreIncompatible -> raise (Incompatible TyBody)
+
+
+  let for_opfix (hyps : hyps) (f1 : opfix) (f2 : opfix) =
+    let rec for_opbranch (s : EcSubst.subst) (ob1 : opbranch) (ob2 : opbranch) =
+      check (EcPath.p_equal (fst ob1.opb_ctor) (fst ob2.opb_ctor));
+      check (snd ob1.opb_ctor = snd ob2.opb_ctor);
+      for_opbranches hyps s ob1.opb_sub ob2.opb_sub
+
+    and for_opbranches (hyps : hyps) (s : EcSubst.subst) (ob1 : opbranches) (ob2 : opbranches) =
+      match ob1, ob2 with
+      | OPB_Leaf (d1, e1), OPB_Leaf (d2, e2) ->
+        check (List.compare_lengths d1 d2 = 0);
+        let s = List.fold_left2 (for_params hyps) s d1 d2 in
+        for_expr hyps s e1 e2
+
+      | OPB_Branch obs1, OPB_Branch obs2 ->
+        check (Parray.length obs1 = Parray.length obs2);
+        Parray.iter2 (for_opbranch s) obs1 obs2
+
+      | _, _ -> raise CoreIncompatible
+      in
+
+    check (EcReduction.EqTest.for_type (toenv hyps) f1.opf_resty f2.opf_resty);
+    check (f1.opf_struct = f2.opf_struct);
+
+    let s = for_params hyps EcSubst.empty f1.opf_args f2.opf_args in
+    let s = EcSubst.add_path  ~src:f2.opf_recp ~dst:f1.opf_recp s
+
+    in for_opbranches hyps s f1.opf_branches f2.opf_branches
+
+  let for_ind (hyps : hyps) (pi1 : prind) (pi2 : prind) =
+    let for_prctor (s : EcSubst.subst) (prc1 : prctor) (prc2 : prctor) =
+      check (EcSymbols.sym_equal prc1.prc_ctor prc2.prc_ctor);
+      let (env, s) =
+        EcReduction.check_bindings
+          CoreIncompatible (toenv hyps) s prc1.prc_bds prc2.prc_bds in
+      let hyps = EcEnv.LDecl.init env [] in
+      check (List.compare_lengths prc1.prc_spec prc2.prc_spec = 0);
+      let for_spec (f1 : form) (f2 : form) =
+        check (EcReduction.is_conv hyps f1 (EcSubst.subst_form s f2)) in
+      List.iter2 for_spec prc1.prc_spec prc2.prc_spec
+    in
+
+    let s = for_params hyps EcSubst.empty pi1.pri_args pi2.pri_args in
+    check (List.compare_lengths pi1.pri_ctors pi2.pri_ctors = 0);
+    List.iter2 (for_prctor s) pi1.pri_ctors pi2.pri_ctors
+
+  let rec for_oper (hyps : hyps) (ob1 : opbody) (ob2 : opbody) =
+    match ob1, ob2 with
+    | OP_Plain f1, OP_Plain f2 ->
+      check (EcReduction.is_conv ~ri:ri_compatible hyps f1 f2)
+
+    | OP_Plain { f_node = Fop (p, tys) }, _ ->
+      let ob1 = get_open_oper (toenv hyps) p tys  in
+      for_oper hyps ob1 ob2
+
+    | _, OP_Plain { f_node = Fop (p, tys) } ->
+      let ob2 = get_open_oper (toenv hyps) p tys in
+      for_oper hyps ob1 ob2
+
+    | OP_Constr (p1, i1), OP_Constr (p2, i2) ->
+      check (EcPath.p_equal p1 p2 && i1 = i2)
+
+    | OP_Record p1, OP_Record p2 ->
+      check (EcPath.p_equal p1 p2)
+
+    | OP_Proj (p1, i11, i12), OP_Proj (p2, i21, i22) ->
+      check (EcPath.p_equal p1 p2 && i11 = i21 && i12 = i22)
+
+    | OP_Fix f1, OP_Fix f2 ->
+      for_opfix hyps f1 f2
+
+    | OP_TC, OP_TC -> ()
+
+    | _, _ -> raise CoreIncompatible
+
+  let rec for_pred (hyps : EcEnv.LDecl.hyps) (pb1 : prbody) (pb2 : prbody) =
+    match pb1, pb2 with
+    | PR_Plain f1, PR_Plain f2 ->
+      check (EcReduction.is_conv hyps f1 f2)
+
+    | PR_Plain { f_node = Fop (p, tys) }, _ ->
+      let pb1 = get_open_pred (toenv hyps) p tys  in
+      for_pred hyps pb1 pb2
+
+    | _, PR_Plain { f_node = Fop (p, tys) } ->
+      let pb2 = get_open_pred (toenv hyps) p tys  in
+      for_pred hyps pb1 pb2
+
+    | PR_Ind pr1, PR_Ind pr2 ->
+      for_ind hyps pr1 pr2
+
+    | _, _ -> raise CoreIncompatible
+
+  let for_nott (hyps : hyps) (nb1 : notation) (nb2 : notation) =
+    let s = for_params hyps EcSubst.empty nb1.ont_args nb2.ont_args in
+    (* We do not check ont_resty because it is redundant *)
+    for_expr hyps s nb1.ont_body nb2.ont_body
+
+  let for_operator (env : EcEnv.env) (oper1 : operator) (oper2 : operator) =
+    let params = oper1.op_tparams in
+
+    for_tparams oper1.op_tparams oper2.op_tparams;
+
+    let oty1, okind1 = oper1.op_ty, oper1.op_kind in
+    let tparams = List.map (fun (id,_) -> tvar id) params in
+    let oty2, okind2 = EcSubst.open_oper oper2 tparams in
+
+    if not (EcReduction.EqTest.for_type env oty1 oty2) then
+      raise (Incompatible (DifferentType(oty1, oty2)));
+
+    let hyps = EcEnv.LDecl.init env params in
+
+    try
+      match okind1, okind2 with
+      | OB_oper None      , OB_oper _          -> ()
+      | OB_oper (Some ob1), OB_oper (Some ob2) -> for_oper hyps ob2 ob1
+      | OB_pred None      , OB_pred _          -> ()
+      | OB_pred (Some pb1), OB_pred (Some pb2) -> for_pred hyps pb2 pb1
+      | OB_nott nb1       , OB_nott nb2        -> for_nott hyps nb2 nb1
+      | _                 , _                  -> raise (Incompatible OpBody)
+
+    with Failure _ -> raise (Incompatible OpBody)
+end
 
 (* -------------------------------------------------------------------- *)
 let check_evtags ?(tags : evtags option) (src : symbol list) =
@@ -425,7 +496,7 @@ let rec replay_tyd (ove : _ ovrenv) (subst, ops, proofs, scope) (import, x, otyd
       let refotyd = EcSubst.subst_tydecl subst otyd in
 
       begin
-        try tydecl_compatible env refotyd newtyd
+        try Compatible.for_tydecl env refotyd newtyd
         with Incompatible err ->
           clone_error env (CE_TyIncompatible ((snd ove.ovre_prefix, x), err))
       end;
@@ -476,7 +547,7 @@ and replay_opd (ove : _ ovrenv) (subst, ops, proofs, scope) (import, x, oopd) =
                 (lam.f_ty, lam)
               in
               begin
-                try ty_compatible env ue
+                try Compatible.for_ty env ue
                     (List.map fst reftyvars, refty)
                     (List.map fst (EcUnify.UniEnv.tparams ue), ty)
                 with Incompatible err ->
@@ -533,7 +604,7 @@ and replay_opd (ove : _ ovrenv) (subst, ops, proofs, scope) (import, x, oopd) =
           | `Inline _ ->
               let body =
                 try
-                  EcFol.expr_of_form EcFol.mhr body
+                  EcFol.expr_of_form body
                 with EcFol.CannotTranslate ->
                   clone_error env (CE_InlinedOpIsForm (snd ove.ovre_prefix, x))
               in
@@ -547,7 +618,7 @@ and replay_opd (ove : _ ovrenv) (subst, ops, proofs, scope) (import, x, oopd) =
         Mp.add opp (newop, alias) ops in
 
       begin
-        try operator_compatible env refop newop
+        try Compatible.for_operator env refop newop
         with Incompatible err ->
           clone_error env (CE_OpIncompatible ((snd ove.ovre_prefix, x), err))
       end;
@@ -599,7 +670,7 @@ and replay_prd (ove : _ ovrenv) (subst, ops, proofs, scope) (import, x, oopr) =
 
             begin
               try
-                ty_compatible env ue
+                Compatible.for_ty env ue
                   (List.map fst reftyvars, refty)
                   (List.map fst (EcUnify.UniEnv.tparams ue), body.f_ty)
               with Incompatible err ->
@@ -669,7 +740,7 @@ and replay_prd (ove : _ ovrenv) (subst, ops, proofs, scope) (import, x, oopr) =
     in
 
     begin
-      try operator_compatible env refpr newpr
+      try Compatible.for_operator env refpr newpr
       with Incompatible err ->
         clone_error env (CE_OpIncompatible ((snd ove.ovre_prefix, x), err))
     end;
