@@ -3,6 +3,7 @@ open EcUtils
 open EcTypes
 open EcFol
 open EcEnv
+open EcAst
 
 open EcCoreGoal
 open EcLowGoal
@@ -47,25 +48,26 @@ end
 let t_hoare_cond tc =
   let hs = tc1_as_hoareS tc in
   let (e,_,_) = fst (tc1_first_if tc hs.hs_s) in
-  LowInternal.t_gen_cond None (form_of_expr (EcMemory.memory hs.hs_m) e) tc
+  LowInternal.t_gen_cond None (Inv_ss (ss_inv_of_expr (EcMemory.memory hs.hs_m) e)) tc
 
 (* -------------------------------------------------------------------- *)
 let t_ehoare_cond tc =
   let hs = tc1_as_ehoareS tc in
   let (e,_,_) = fst (tc1_first_if tc hs.ehs_s) in
   LowInternal.t_gen_cond ~t_finalize:LowInternal.t_finalize_ehoare
-    None (form_of_expr (EcMemory.memory hs.ehs_m) e) tc
+    None (Inv_ss (ss_inv_of_expr (EcMemory.memory hs.ehs_m) e)) tc
 
 (* -------------------------------------------------------------------- *)
 let t_bdhoare_cond tc =
   let bhs = tc1_as_bdhoareS tc in
   let (e,_,_) = fst (tc1_first_if tc bhs.bhs_s) in
-  LowInternal.t_gen_cond None (form_of_expr (EcMemory.memory bhs.bhs_m) e) tc
+  LowInternal.t_gen_cond None (Inv_ss (ss_inv_of_expr (EcMemory.memory bhs.bhs_m) e)) tc
 
 (* -------------------------------------------------------------------- *)
 let rec t_equiv_cond side tc =
   let hyps = FApi.tc1_hyps tc in
   let es   = tc1_as_equivS tc in
+  let ml, mr = fst es.es_ml, fst es.es_mr in
 
   match side with
   | Some s ->
@@ -73,21 +75,20 @@ let rec t_equiv_cond side tc =
         match s with
         | `Left ->
           let (e,_,_) = fst (tc1_first_if tc es.es_sl) in
-          form_of_expr (EcMemory.memory es.es_ml) e
+          ss_inv_generalize_right (ss_inv_of_expr ml e) mr
         | `Right ->
           let (e,_,_) = fst (tc1_first_if tc es.es_sr) in
-          form_of_expr (EcMemory.memory es.es_mr) e
-      in LowInternal.t_gen_cond side e tc
+          ss_inv_generalize_left (ss_inv_of_expr mr e) ml
+      in LowInternal.t_gen_cond side (Inv_ts e) tc
 
   | None ->
       let el,_,_ = fst (tc1_first_if tc es.es_sl) in
       let er,_,_ = fst (tc1_first_if tc es.es_sr) in
-      let el     = form_of_expr (EcMemory.memory es.es_ml) el in
-      let er     = form_of_expr (EcMemory.memory es.es_mr) er in
+      let el     = ss_inv_generalize_right (ss_inv_of_expr ml el) mr in
+      let er     = ss_inv_generalize_left (ss_inv_of_expr mr er) ml in
       let fiff   =
-        f_forall_mems
-          [es.es_ml;es.es_mr]
-          (f_imp es.es_pr (f_iff el er)) in
+        EcSubst.f_forall_mems_ts_inv es.es_ml es.es_mr
+          (map_ts_inv2 f_imp (es_pr es) (map_ts_inv2 f_iff el er)) in
 
       let fresh = ["hiff";"&m1";"&m2";"h";"h";"h"] in
       let fresh = LDecl.fresh_ids hyps fresh in
@@ -131,7 +132,17 @@ end = struct
     let (e, _), _ = tc1_first_match tc st in
     let _, indt, _ = oget (EcEnv.Ty.get_top_decl e.e_ty env) in
     let indt = oget (EcDecl.tydecl_as_datatype indt) in
-    let f = form_of_expr (EcMemory.memory me) e in
+    let f = 
+      let f = ss_inv_of_expr (EcMemory.memory me) e in
+      let pre = tc1_get_pre tc in
+      match pre, side with
+      | Inv_ts {mr}, Some `Left ->
+        Inv_ts (ss_inv_generalize_right f mr)
+      | Inv_ts {ml}, Some `Right ->
+        Inv_ts (ss_inv_generalize_left f ml)
+      | Inv_ss _, _ ->
+        Inv_ss f
+      | Inv_ts _, None -> tc_error !!tc "expecting a side" in
 
     let onsub (i : int) (tc : tcenv1) =
       let cname, cargs = List.nth indt.tydt_ctors i in
@@ -200,9 +211,12 @@ end = struct
 
         let pre = oget (EcLowPhlGoal.get_pre (FApi.tc1_goal tc)) in
         let post = oget (EcLowPhlGoal.get_post (FApi.tc1_goal tc)) in
-        let eq, _, pre = destr_and3 pre in
+
+        let pre = map_inv1 (fun pre -> 
+          let eq, _, pre = destr_and3 pre in
+          f_and eq pre) pre in
         tc
-          |> EcPhlConseq.t_conseq (f_and eq pre) post
+          |> EcPhlConseq.t_conseq pre post
           @+ [discharge_pre; discharge_post; EcLowGoal.t_clears names]
       in
 
@@ -238,6 +252,7 @@ let t_equiv_match_same_constr tc =
   let hyps = FApi.tc1_hyps tc in
   let env  = LDecl.toenv hyps in
   let es   = tc1_as_equivS tc in
+  let ml, mr = fst es.es_ml, fst es.es_mr in
 
   let (el, bsl), sl = tc1_first_match tc es.es_sl in
   let (er, bsr), sr = tc1_first_match tc es.es_sr in
@@ -249,44 +264,45 @@ let t_equiv_match_same_constr tc =
     tc_error !!tc "match statements on different inductive types";
 
   let dt = oget (EcDecl.tydecl_as_datatype dt) in
-  let fl = form_of_expr (EcMemory.memory es.es_ml) el in
-  let fr = form_of_expr (EcMemory.memory es.es_mr) er in
+  let fl = ss_inv_generalize_right (ss_inv_of_expr ml el) mr in
+  let fr = ss_inv_generalize_left (ss_inv_of_expr mr er) ml in
 
   let get_eqv_cond ((c, _), ((cl, _), (cr, _))) =
     let bhl  = List.map (fst_map EcIdent.fresh) cl in
     let bhr  = List.map (fst_map EcIdent.fresh) cr in
     let cop  = EcPath.pqoname (EcPath.prefix pl) c in
-    let copl = f_op cop tyl (toarrow (List.snd cl) fl.f_ty) in
-    let copr = f_op cop tyr (toarrow (List.snd cr) fr.f_ty) in
+    let copl = f_op cop tyl (toarrow (List.snd cl) fl.inv.f_ty) in
+    let copr = f_op cop tyr (toarrow (List.snd cr) fr.inv.f_ty) in
 
-    let lhs = f_eq fl (f_app copl (List.map (curry f_local) bhl) fl.f_ty) in
-    let lhs = f_exists (List.map (snd_map gtty) bhl) lhs in
+    let lhs = map_ts_inv1 (fun fl -> f_eq fl (f_app copl (List.map (curry f_local) bhl) fl.f_ty)) fl in
+    let lhs = map_ts_inv1 (f_exists (List.map (snd_map gtty) bhl)) lhs in
 
-    let rhs = f_eq fr (f_app copr (List.map (curry f_local) bhr) fr.f_ty) in
-    let rhs = f_exists (List.map (snd_map gtty) bhr) rhs in
+    let rhs = map_ts_inv1 (fun fr -> f_eq fr (f_app copr (List.map (curry f_local) bhr) fr.f_ty)) fr in
+    let rhs = map_ts_inv1 (f_exists (List.map (snd_map gtty) bhr)) rhs in
 
-    f_forall_mems [es.es_ml; es.es_mr] (f_imp_simpl es.es_pr (f_iff lhs rhs)) in
+    EcSubst.f_forall_mems_ts_inv es.es_ml es.es_mr 
+      (map_ts_inv2 f_imp_simpl (es_pr es) (map_ts_inv2 f_iff lhs rhs)) in
 
   let get_eqv_goal ((c, _), ((cl, bl), (cr, br))) =
     let sb      = Fsubst.f_subst_id in
     let sb, bhl = add_elocals sb cl in
     let sb, bhr = add_elocals sb cr in
     let cop     = EcPath.pqoname (EcPath.prefix pl) c in
-    let copl    = f_op cop tyl (toarrow (List.snd cl) fl.f_ty) in
-    let copr    = f_op cop tyr (toarrow (List.snd cr) fr.f_ty) in
-    let pre     = f_ands_simpl
-      [ f_eq fl (f_app copl (List.map (curry f_local) bhl) fl.f_ty);
-        f_eq fr (f_app copr (List.map (curry f_local) bhr) fr.f_ty) ]
-      es.es_pr in
+    let copl    = f_op cop tyl (toarrow (List.snd cl) fl.inv.f_ty) in
+    let copr    = f_op cop tyr (toarrow (List.snd cr) fr.inv.f_ty) in
+    let f_ands_simpl' f = f_ands_simpl (List.tl f) (List.hd f) in
+    let pre     = map_ts_inv f_ands_simpl'
+      [es_pr es; map_ts_inv1 (fun fl -> f_eq fl (f_app copl (List.map (curry f_local) bhl) fl.f_ty)) fl;
+        map_ts_inv1 (fun fr -> f_eq fr (f_app copr (List.map (curry f_local) bhr) fr.f_ty)) fr ]
+       in
 
     f_forall
       ( (List.map (snd_map gtty) bhl) @
         (List.map (snd_map gtty) bhr) )
-      ( f_equivS_r
-          { es with
-              es_sl = EcModules.stmt ((s_subst sb bl).s_node @ sl.s_node);
-              es_sr = EcModules.stmt ((s_subst sb br).s_node @ sr.s_node);
-              es_pr = pre; } )
+      ( f_equivS (snd es.es_ml) (snd es.es_mr) pre 
+        (EcModules.stmt ((s_subst sb bl).s_node @ sl.s_node)) 
+        (EcModules.stmt ((s_subst sb br).s_node @ sr.s_node))
+        (es_po es))
 
   in
 
@@ -303,6 +319,7 @@ let t_equiv_match_eq tc =
   let hyps = FApi.tc1_hyps tc in
   let env  = LDecl.toenv hyps in
   let es   = tc1_as_equivS tc in
+  let ml, mr = fst es.es_ml, fst es.es_mr in
 
   let (el, bsl), sl = tc1_first_match tc es.es_sl in
   let (er, bsr), sr = tc1_first_match tc es.es_sr in
@@ -317,12 +334,12 @@ let t_equiv_match_eq tc =
     tc_error !!tc "synced match requires matches on the same type";
 
   let dt = oget (EcDecl.tydecl_as_datatype dt) in
-  let fl = form_of_expr (EcMemory.memory es.es_ml) el in
-  let fr = form_of_expr (EcMemory.memory es.es_mr) er in
+  let fl = ss_inv_generalize_right (ss_inv_of_expr ml el) mr in
+  let fr = ss_inv_generalize_left (ss_inv_of_expr mr er) ml in
 
   let eqv_cond =
-    f_forall_mems [es.es_ml; es.es_mr]
-      (f_imp_simpl es.es_pr (f_eq fl fr)) in
+    EcSubst.f_forall_mems_ts_inv es.es_ml es.es_mr
+      (map_ts_inv2 f_imp_simpl (es_pr es) (map_ts_inv2 f_eq fl fr)) in
 
   let get_eqv_goal ((c, _), ((cl, bl), (cr, br))) =
     let sb     = f_subst_init () in
@@ -335,20 +352,19 @@ let t_equiv_match_eq tc =
         sb cl cr in
 
     let cop    = EcPath.pqoname (EcPath.prefix pl) c in
-    let copl   = f_op cop tyl (toarrow (List.snd cl) fl.f_ty) in
-    let copr   = f_op cop tyr (toarrow (List.snd cr) fr.f_ty) in
-    let pre    = f_ands_simpl
-      [ f_eq fl (f_app copl (List.map (curry f_local) bh) fl.f_ty);
-        f_eq fr (f_app copr (List.map (curry f_local) bh) fr.f_ty) ]
-      es.es_pr in
+    let copl   = f_op cop tyl (toarrow (List.snd cl) fl.inv.f_ty) in
+    let copr   = f_op cop tyr (toarrow (List.snd cr) fr.inv.f_ty) in
+    let f_ands_simpl' f = f_ands_simpl (List.tl f) (List.hd f) in
+    let pre    = map_ts_inv f_ands_simpl'
+      [ es_pr es; map_ts_inv1 (fun fl -> f_eq fl (f_app copl (List.map (curry f_local) bh) fl.f_ty)) fl;
+        map_ts_inv1 (fun fr -> f_eq fr (f_app copr (List.map (curry f_local) bh) fr.f_ty)) fr ] in
 
     f_forall
       (List.map (snd_map gtty) bh)
-      (f_equivS_r
-         { es with
-             es_sl = EcModules.stmt ((s_subst sb bl).s_node @ sl.s_node);
-             es_sr = EcModules.stmt ((s_subst sb br).s_node @ sr.s_node);
-             es_pr = pre; } )
+      (f_equivS (snd es.es_ml) (snd es.es_mr) pre
+        (EcModules.stmt ((s_subst sb bl).s_node @ sl.s_node))
+        (EcModules.stmt ((s_subst sb br).s_node @ sr.s_node))
+        (es_po es))
 
   in
 
