@@ -1,12 +1,12 @@
 (* -------------------------------------------------------------------- *)
 open EcParsetree
-open EcFol
 open EcCoreGoal.FApi
+open EcAst
 
 (* -------------------------------------------------------------------- *)
 val t_hr_exists_elim_r : ?bound:int -> backward
 val t_hr_exists_elim   : backward
-val t_hr_exists_intro  : form list -> backward
+val t_hr_exists_intro  : inv list -> backward
 
 (* -------------------------------------------------------------------- *)
 val process_exists_intro : elim:bool -> pformula list -> backward
