@@ -1,16 +1,16 @@
 (* -------------------------------------------------------------------- *)
 open EcPath
 open EcParsetree
-open EcFol
+open EcAst
 open EcCoreGoal.FApi
 open EcMatching.Position
 
 (* -------------------------------------------------------------------- *)
 val t_failure_event :
      codepos1
-  -> form -> form -> form -> form
-  -> (xpath * form) list
-  -> form
+  -> ss_inv -> form -> form -> ss_inv
+  -> (xpath * ss_inv) list
+  -> ss_inv
   -> backward
 
 (* -------------------------------------------------------------------- *)
