@@ -9,6 +9,7 @@ open EcMemory
 open EcModules
 open EcEnv
 open EcPV
+open EcSubst
 
 open EcCoreGoal
 open EcLowPhlGoal
@@ -67,7 +68,7 @@ let subst_pre env fs (m : memory) s =
     | Some v -> { v_name = v; v_type = ov.ov_type }
   in
   let v = List.map (fun v -> f_pvloc (fresh v) m) fs.fs_anames in
-  PVM.add env pv_arg m (f_tuple v) s
+  PVM.add env pv_arg m (map_ss_inv ~m f_tuple v).inv s
 
 (* ------------------------------------------------------------------ *)
 let t_hoareF_fun_def_r tc =
@@ -75,12 +76,12 @@ let t_hoareF_fun_def_r tc =
   let hf = tc1_as_hoareF tc in
   let f = NormMp.norm_xfun env hf.hf_f in
   check_concrete !!tc env f;
-  let (memenv, (fsig, fdef), env) = Fun.hoareS f env in
+  let (memenv, (fsig, fdef), env) = Fun.hoareS hf.hf_m f env in
   let m = EcMemory.memory memenv in
-  let fres = odfl f_tt (omap (form_of_expr m) fdef.f_ret) in
-  let post = PVM.subst1 env pv_res m fres hf.hf_po in
-  let pre  = PVM.subst env (subst_pre env fsig m PVM.empty) hf.hf_pr in
-  let concl' = f_hoareS memenv pre fdef.f_body post in
+  let fres = odfl {m;inv=f_tt} (omap (ss_inv_of_expr m) fdef.f_ret) in
+  let post = map_ss_inv2 (PVM.subst1 env pv_res m) fres (hf_po hf) in
+  let pre  = map_ss_inv1 (PVM.subst env (subst_pre env fsig m PVM.empty)) (hf_pr hf) in
+  let concl' = f_hoareS (snd memenv) pre fdef.f_body post in
   FApi.xmutate1 tc `FunDef [concl']
 
 (* ------------------------------------------------------------------ *)
@@ -89,12 +90,12 @@ let t_ehoareF_fun_def_r tc =
   let hf = tc1_as_ehoareF tc in
   let f = NormMp.norm_xfun env hf.ehf_f in
   check_concrete !!tc env f;
-  let (memenv, (fsig, fdef), env) = Fun.hoareS f env in
+  let (memenv, (fsig, fdef), env) = Fun.hoareS hf.ehf_m f env in
   let m = EcMemory.memory memenv in
-  let fres  = odfl f_tt (omap (form_of_expr m) fdef.f_ret) in
-  let post  = PVM.subst1 env pv_res m fres hf.ehf_po in
-  let pre   = PVM.subst env (subst_pre env fsig m PVM.empty) hf.ehf_pr in
-  let concl' = f_eHoareS memenv pre fdef.f_body post in
+  let fres  = odfl {m;inv=f_tt} (omap (ss_inv_of_expr m) fdef.f_ret) in
+  let post  = map_ss_inv2 (PVM.subst1 env pv_res m) fres (ehf_po hf) in
+  let pre   = map_ss_inv1 (PVM.subst env (subst_pre env fsig m PVM.empty)) (ehf_pr hf) in
+  let concl' = f_eHoareS (snd memenv) pre fdef.f_body post in
   FApi.xmutate1 tc `FunDef [concl']
 
 (* ------------------------------------------------------------------ *)
@@ -103,37 +104,36 @@ let t_bdhoareF_fun_def_r tc =
   let bhf = tc1_as_bdhoareF tc in
   let f = NormMp.norm_xfun env bhf.bhf_f in
   check_concrete !!tc env f;
-  let (memenv, (fsig, fdef), env) = Fun.hoareS f env in
+  let (memenv, (fsig, fdef), env) = Fun.hoareS bhf.bhf_m f env in
   let m = EcMemory.memory memenv in
-  let fres = odfl f_tt (omap (form_of_expr m) fdef.f_ret) in
-  let post = PVM.subst1 env pv_res m fres bhf.bhf_po in
+  let fres = odfl {m;inv=f_tt} (omap (ss_inv_of_expr m) fdef.f_ret) in
+  let post = map_ss_inv2 (PVM.subst1 env pv_res m) fres (bhf_po bhf) in
   let spre = subst_pre env fsig m PVM.empty in
-  let pre = PVM.subst env spre bhf.bhf_pr in
-  let bd  = PVM.subst env spre bhf.bhf_bd in
-  let concl' = f_bdHoareS memenv pre fdef.f_body post bhf.bhf_cmp bd in
+  let pre = map_ss_inv1 (PVM.subst env spre) (bhf_pr bhf) in
+  let bd  = map_ss_inv1 (PVM.subst env spre) (bhf_bd bhf) in
+  let concl' = f_bdHoareS (snd memenv) pre fdef.f_body post bhf.bhf_cmp bd in
   FApi.xmutate1 tc `FunDef [concl']
 
 (* ------------------------------------------------------------------ *)
 let t_equivF_fun_def_r tc =
   let env = FApi.tc1_env tc in
   let ef = tc1_as_equivF tc in
+  let ml, mr = ef.ef_ml, ef.ef_mr in
   let fl = NormMp.norm_xfun env ef.ef_fl in
   let fr = NormMp.norm_xfun env ef.ef_fr in
   check_concrete !!tc env fl; check_concrete !!tc env fr;
-  let (menvl, eqsl, menvr, eqsr, env) = Fun.equivS fl fr env in
+  let (menvl, eqsl, menvr, eqsr, env) = Fun.equivS ml mr fl fr env in
   let (fsigl, fdefl) = eqsl in
   let (fsigr, fdefr) = eqsr in
-  let ml = EcMemory.memory menvl in
-  let mr = EcMemory.memory menvr in
-  let fresl = odfl f_tt (omap (form_of_expr ml) fdefl.f_ret) in
-  let fresr = odfl f_tt (omap (form_of_expr mr) fdefr.f_ret) in
-  let s = PVM.add env pv_res ml fresl PVM.empty in
-  let s = PVM.add env pv_res mr fresr s in
-  let post = PVM.subst env s ef.ef_po in
+  let fresl = odfl {m=ml;inv=f_tt} (omap (ss_inv_of_expr ml) fdefl.f_ret) in
+  let fresr = odfl {m=mr;inv=f_tt} (omap (ss_inv_of_expr mr) fdefr.f_ret) in
+  let s = PVM.add env pv_res ml fresl.inv PVM.empty in
+  let s = PVM.add env pv_res mr fresr.inv s in
+  let post = map_ts_inv1 (PVM.subst env s) (ef_po ef) in
   let s = subst_pre env fsigl ml PVM.empty in
   let s = subst_pre env fsigr mr s in
-  let pre = PVM.subst env s ef.ef_pr in
-  let concl' = f_equivS menvl menvr pre fdefl.f_body fdefr.f_body post in
+  let pre = map_ts_inv1 (PVM.subst env s) (ef_pr ef) in
+  let concl' = f_equivS (snd menvl) (snd menvr) pre fdefl.f_body fdefr.f_body post in
   FApi.xmutate1 tc `FunDef [concl']
 
 (* -------------------------------------------------------------------- *)
@@ -158,16 +158,16 @@ module FunAbsLow = struct
   (* ------------------------------------------------------------------ *)
   let hoareF_abs_spec _pf env f inv =
     let (top, _, oi, _) = EcLowPhlGoal.abstract_info env f in
-    let fv = PV.fv env mhr inv in
+    let fv = PV.fv env inv.m inv.inv in
     PV.check_depend env fv top;
     let ospec o = f_hoareF inv o inv in
     let sg = List.map ospec (OI.allowed oi) in
     (inv, inv, sg)
 
   (* ------------------------------------------------------------------ *)
-  let ehoareF_abs_spec _pf env f inv =
+  let ehoareF_abs_spec _pf env f (inv: ss_inv) =
     let (top, _, oi, _) = EcLowPhlGoal.abstract_info env f in
-    let fv = PV.fv env mhr inv in
+    let fv = PV.fv env inv.m inv.inv in
     PV.check_depend env fv top;
     let ospec o = f_eHoareF inv o inv in
     let sg = List.map ospec (OI.allowed oi) in
@@ -176,28 +176,28 @@ module FunAbsLow = struct
   (* ------------------------------------------------------------------ *)
   let bdhoareF_abs_spec pf env f inv =
     let (top, _, oi, _) = EcLowPhlGoal.abstract_info env f in
-    let fv = PV.fv env mhr inv in
+    let fv = PV.fv env inv.m inv.inv in
 
     PV.check_depend env fv top;
     let ospec o =
       check_oracle_use pf env top o;
-      f_bdHoareF inv o inv FHeq f_r1 in
+      f_bdHoareF inv o inv FHeq {m=inv.m;inv=f_r1} in
 
     let sg = List.map ospec (OI.allowed oi) in
     (inv, inv, lossless_hyps env top f.x_sub :: sg)
 
   (* ------------------------------------------------------------------ *)
-  let equivF_abs_spec pf env fl fr inv =
+  let equivF_abs_spec pf env fl fr (inv: ts_inv) =
     let (topl, _fl, oil, sigl), (topr, _fr, oir, sigr) =
       EcLowPhlGoal.abstract_info2 env fl fr
     in
 
-    let ml, mr = mleft, mright in
-    let fvl = PV.fv env ml inv in
-    let fvr = PV.fv env mr inv in
+    let ml, mr = inv.ml, inv.mr in
+    let fvl = PV.fv env inv.ml inv.inv in
+    let fvr = PV.fv env inv.mr inv.inv in
     PV.check_depend env fvl topl;
     PV.check_depend env fvr topr;
-    let eqglob = f_eqglob topl ml topr mr in
+    let eqglob = ts_inv_eqglob topl ml topr mr in
 
     let ospec o_l o_r =
       let use =
@@ -221,8 +221,8 @@ module FunAbsLow = struct
         f_eqres fo_l.f_sig.fs_ret ml fo_r.f_sig.fs_ret mr in
 
       let invs = if use then [eqglob; inv] else [inv] in
-      let pre  = EcFol.f_ands (eq_params :: invs) in
-      let post = EcFol.f_ands (eq_res :: invs) in
+      let pre  = map_ts_inv (fun invs -> EcFol.f_ands (eq_params :: invs)) invs in
+      let post = map_ts_inv (fun invs -> EcFol.f_ands (eq_res :: invs)) invs in
       f_equivF pre o_l o_r post
     in
 
@@ -233,10 +233,10 @@ module FunAbsLow = struct
         sigl.fs_arg sigl.fs_anames ml
         sigr.fs_arg sigr.fs_anames mr in
 
-    let eq_res = f_eqres sigl.fs_ret ml sigr.fs_ret mr in
+    let eq_res = ts_inv_eqres sigl.fs_ret ml sigr.fs_ret mr in
     let lpre   = [eqglob;inv] in
-    let pre    = f_ands (eq_params::lpre) in
-    let post   = f_ands [eq_res; eqglob; inv] in
+    let pre    = map_ts_inv (fun lpre -> f_ands (eq_params::lpre)) lpre in
+    let post   = map_ts_inv f_ands [eq_res; eqglob; inv] in
 
     (pre, post, sg)
 end
@@ -264,7 +264,7 @@ let t_bdhoareF_abs_r inv tc =
   let bhf = tc1_as_bdhoareF tc in
 
   match bhf.bhf_cmp with
-  | FHeq when f_equal bhf.bhf_bd f_r1 ->
+  | FHeq when f_equal (bhf_bd bhf).inv f_r1 ->
       let pre, post, sg =
         FunAbsLow.bdhoareF_abs_spec !!tc env bhf.bhf_f inv
       in
@@ -282,7 +282,7 @@ let t_equivF_abs_r inv tc =
   in
 
   let tactic tc = FApi.xmutate1 tc `FunAbs sg in
-  FApi.t_last tactic (EcPhlConseq.t_equivF_conseq pre post tc)
+    FApi.t_last tactic (EcPhlConseq.t_equivF_conseq pre post tc)
 
 (* -------------------------------------------------------------------- *)
 let t_hoareF_abs   = FApi.t_low1 "hoare-fun-abs"   t_hoareF_abs_r
@@ -293,22 +293,23 @@ let t_equivF_abs   = FApi.t_low1 "equiv-fun-abs"   t_equivF_abs_r
 (* -------------------------------------------------------------------- *)
 module UpToLow = struct
   (* ------------------------------------------------------------------ *)
-  let equivF_abs_upto pf env fl fr bad invP invQ =
+  let equivF_abs_upto pf env fl fr (bad: ss_inv) (invP: ts_inv) (invQ: ts_inv) =
     let (topl, _fl, oil, sigl), (topr, _fr, oir, sigr) =
       EcLowPhlGoal.abstract_info2 env fl fr
     in
 
-    let ml, mr = mleft, mright in
-    let bad2 = Fsubst.f_subst_mem mhr mr bad in
-    let allinv = f_ands [bad2; invP; invQ] in
-    let fvl = PV.fv env ml allinv in
-    let fvr = PV.fv env mr allinv in
+    let ml, mr = invP.ml, invP.mr in
+    let bad = ss_inv_rebind bad invP.mr in
+    let bad2 = ss_inv_generalize_left bad invP.ml in
+    let allinv = map_ts_inv f_ands [bad2; invP; invQ] in
+    let fvl = PV.fv env ml allinv.inv in
+    let fvr = PV.fv env mr allinv.inv in
 
     PV.check_depend env fvl topl;
     PV.check_depend env fvr topr;
 
     (* FIXME: check there is only global variable *)
-    let eqglob = f_eqglob topl ml topr mr in
+    let eqglob = ts_inv_eqglob topl ml topr mr in
 
     let ospec o_l o_r =
       check_oracle_use pf env topl o_l;
@@ -318,25 +319,25 @@ module UpToLow = struct
       let fo_l = EcEnv.Fun.by_xpath o_l env in
       let fo_r = EcEnv.Fun.by_xpath o_r env in
       let eq_params =
-        f_eqparams
+        ts_inv_eqparams
           fo_l.f_sig.fs_arg fo_l.f_sig.fs_anames ml
           fo_r.f_sig.fs_arg fo_r.f_sig.fs_anames mr in
 
       let eq_res =
-        f_eqres fo_l.f_sig.fs_ret ml fo_r.f_sig.fs_ret mr in
+        ts_inv_eqres fo_l.f_sig.fs_ret ml fo_r.f_sig.fs_ret mr in
 
-      let pre   = EcFol.f_ands [EcFol.f_not bad2; eq_params; invP] in
-      let post  = EcFol.f_if_simpl bad2 invQ (f_and eq_res invP) in
+      let pre   = map_ts_inv EcFol.f_ands [map_ts_inv1 EcFol.f_not bad2; eq_params; invP] in
+      let post  = map_ts_inv3 EcFol.f_if_simpl bad2 invQ (map_ts_inv2 f_and eq_res invP) in
       let cond1 = f_equivF pre o_l o_r post in
       let cond2 =
-        let q = Fsubst.f_subst_mem ml EcFol.mhr invQ in
-          f_forall[(mr, GTmem abstract_mt)]
-            (f_imp bad2 (f_bdHoareF q o_l q FHeq f_r1)) in
+        let f_r1 = {m=invQ.ml; inv=f_r1} in
+        let concl = ts_inv_lower_left1 (fun bq -> (f_bdHoareF bq o_l bq FHeq f_r1)) invQ in
+        f_forall_mems_ss_inv (mr, abstract_mt)
+          (map_ss_inv2 f_imp bad concl) in
       let cond3 =
-        let q  = Fsubst.f_subst_mem mr EcFol.mhr invQ in
-        let bq = f_and bad q in
-          f_forall [(ml, GTmem abstract_mt)]
-            (f_bdHoareF bq o_r bq FHeq f_r1) in
+        let f_r1 = {m=invQ.mr; inv=f_r1} in
+        let bq = map_ts_inv2 f_and bad2 invQ in
+          f_forall_mems_ss_inv (ml, abstract_mt) (ts_inv_lower_right1 (fun bq -> f_bdHoareF bq o_r bq FHeq f_r1) bq) in
 
       [cond1; cond2; cond3]
     in
@@ -347,15 +348,15 @@ module UpToLow = struct
     let sg = lossless_a :: sg in
 
     let eq_params =
-      f_eqparams
+      ts_inv_eqparams
         sigl.fs_arg sigl.fs_anames ml
         sigr.fs_arg sigr.fs_anames mr in
 
-    let eq_res = f_eqres sigl.fs_ret ml sigr.fs_ret mr in
+    let eq_res = ts_inv_eqres sigl.fs_ret ml sigr.fs_ret mr in
 
     let pre  = [eqglob;invP] in
-    let pre  = f_if_simpl bad2 invQ (f_ands (eq_params::pre)) in
-    let post = f_if_simpl bad2 invQ (f_ands [eq_res;eqglob;invP]) in
+    let pre  = map_ts_inv3 f_if_simpl bad2 invQ (map_ts_inv f_ands (eq_params::pre)) in
+    let post = map_ts_inv3 f_if_simpl bad2 invQ (map_ts_inv f_ands [eq_res;eqglob;invP]) in
 
     (pre, post, sg)
 end
@@ -404,13 +405,13 @@ module ToCodeLow = struct
       i_call (Some (LvVar (pv_loc (oget res.ov_name), res.ov_type)), f, eargs)
     in (me, stmt [icall], res, args)
 
-  let add_var env vfrom mfrom v me s =
-    PVM.add env vfrom mfrom (f_pvar (pv_loc (oget v.ov_name)) v.ov_type (fst me)) s
+  let add_var env vfrom mfrom v m s =
+    PVM.add env vfrom mfrom (f_pvar (pv_loc (oget v.ov_name)) v.ov_type m).inv s
 
-  let add_var_tuple env vfrom mfrom vs me s =
+  let add_var_tuple env vfrom mfrom vs m s =
     let vs =
-      List.map (fun v -> f_pvar (pv_loc v.v_name) v.v_type (fst me)) vs
-    in PVM.add env vfrom mfrom (f_tuple vs) s
+      List.map (fun v -> f_pvar (pv_loc v.v_name) v.v_type m) vs
+    in PVM.add env vfrom mfrom (map_ss_inv ~m f_tuple vs).inv s
 end
 
 (* -------------------------------------------------------------------- *)
@@ -418,13 +419,14 @@ let t_fun_to_code_hoare_r tc =
   let env = FApi.tc1_env tc in
   let hf = tc1_as_hoareF tc in
   let f = hf.hf_f in
-  let m, st, r, a = ToCodeLow.to_code env f mhr in
-  let spr = ToCodeLow.add_var_tuple env pv_arg mhr a m PVM.empty in
-  let spo = ToCodeLow.add_var env pv_res mhr r m PVM.empty in
-  let pre  = PVM.subst env spr hf.hf_pr in
-  let post = PVM.subst env spo hf.hf_po in
-  let concl = f_hoareS m pre st post in
-
+  let m = hf.hf_m in
+  let (m0, mt), st, r, a = ToCodeLow.to_code env f m in
+  assert (EcIdent.id_equal m0 m);
+  let spr = ToCodeLow.add_var_tuple env pv_arg m a m PVM.empty in
+  let spo = ToCodeLow.add_var env pv_res m r m PVM.empty in
+  let pre  = PVM.subst env spr (hf_pr hf).inv in
+  let post = PVM.subst env spo (hf_po hf).inv in
+  let concl = f_hoareS mt {m;inv=pre} st {m;inv=post} in
   FApi.xmutate1 tc `FunToCode [concl]
 
 (* -------------------------------------------------------------------- *)
@@ -432,15 +434,15 @@ let t_fun_to_code_ehoare_r tc =
   let env = FApi.tc1_env tc in
   let hf = tc1_as_ehoareF tc in
   let f = hf.ehf_f in
-  let m, st, r, a = ToCodeLow.to_code env f mhr in
-  let spr = ToCodeLow.add_var_tuple env pv_arg mhr a m PVM.empty in
-  let spo = ToCodeLow.add_var env pv_res mhr r m PVM.empty in
+  let m = hf.ehf_m in
+  let (m0, mt), st, r, a = ToCodeLow.to_code env f m in
+  assert (EcIdent.id_equal m0 m);
+  let spr = ToCodeLow.add_var_tuple env pv_arg m a m PVM.empty in
+  let spo = ToCodeLow.add_var env pv_res m r m PVM.empty in
 
-  let pre = PVM.subst env spr hf.ehf_pr in
-
-  let post = PVM.subst env spo hf.ehf_po in
-
-  let concl = f_eHoareS m pre st post in
+  let pre = PVM.subst env spr (ehf_pr hf).inv in
+  let post = PVM.subst env spo (ehf_po hf).inv in
+  let concl = f_eHoareS mt {m;inv=pre} st {m;inv=post} in
 
   FApi.xmutate1 tc `FunToCode [concl]
 
@@ -448,59 +450,67 @@ let t_fun_to_code_ehoare_r tc =
 let t_fun_to_code_bdhoare_r tc =
   let env = FApi.tc1_env tc in
   let hf = tc1_as_bdhoareF tc in
+  let m = hf.bhf_m in
   let f = hf.bhf_f in
-  let m, st, r, a = ToCodeLow.to_code env f mhr in
-  let spr = ToCodeLow.add_var_tuple env pv_arg mhr a m PVM.empty in
-  let spo = ToCodeLow.add_var env pv_res mhr r m PVM.empty in
-  let pre  = PVM.subst env spr hf.bhf_pr in
-  let post = PVM.subst env spo hf.bhf_po in
-  let bd   = PVM.subst env spr hf.bhf_bd in
-  let concl = f_bdHoareS m pre st post hf.bhf_cmp bd in
+  let (m0, mt), st, r, a = ToCodeLow.to_code env f m in
+  assert (EcIdent.id_equal m0 m);
+  let spr = ToCodeLow.add_var_tuple env pv_arg m0 a m PVM.empty in
+  let spo = ToCodeLow.add_var env pv_res m0 r m PVM.empty in
+  let pre  = PVM.subst env spr (bhf_pr hf).inv in
+  let post = PVM.subst env spo (bhf_po hf).inv in
+  let bd   = PVM.subst env spr (bhf_bd hf).inv in
+  let concl = f_bdHoareS mt {m;inv=pre} st {m;inv=post} hf.bhf_cmp {m;inv=bd} in
   FApi.xmutate1 tc `FunToCode [concl]
 
 (* -------------------------------------------------------------------- *)
 let t_fun_to_code_equiv_r tc =
   let env = FApi.tc1_env tc in
   let ef = tc1_as_equivF tc in
+  let ml, mr = ef.ef_ml, ef.ef_mr in
   let (fl,fr) = ef.ef_fl, ef.ef_fr in
-  let ml, sl, rl, al = ToCodeLow.to_code env fl mleft in
-  let mr, sr, rr, ar = ToCodeLow.to_code env fr mright in
+  let (ml0, mlt), sl, rl, al = ToCodeLow.to_code env fl ml in
+  assert (EcIdent.id_equal ml0 ml);
+  let (mr0, mrt), sr, rr, ar = ToCodeLow.to_code env fr mr in
+  assert (EcIdent.id_equal mr0 mr);
   let spr =
     let s = PVM.empty in
-    let s = ToCodeLow.add_var_tuple env pv_arg mleft  al ml s in
-    let s = ToCodeLow.add_var_tuple env pv_arg mright ar mr s in
+    let s = ToCodeLow.add_var_tuple env pv_arg ml al ml s in
+    let s = ToCodeLow.add_var_tuple env pv_arg mr ar mr s in
     s in
   let spo =
     let s = PVM.empty in
-    let s = ToCodeLow.add_var env pv_res mleft  rl ml s in
-    let s = ToCodeLow.add_var env pv_res mright rr mr s in
+    let s = ToCodeLow.add_var env pv_res ml rl ml s in
+    let s = ToCodeLow.add_var env pv_res mr rr mr s in
     s in
-  let pre   = PVM.subst env spr ef.ef_pr in
-  let post  = PVM.subst env spo ef.ef_po in
-  let concl = f_equivS ml mr pre sl sr post in
+  let pre   = PVM.subst env spr (ef_pr ef).inv in
+  let post  = PVM.subst env spo (ef_po ef).inv in
+  let concl = f_equivS mlt mrt {ml;mr;inv=pre} sl sr {ml;mr;inv=post} in
 
   FApi.xmutate1 tc `FunToCode [concl]
 
 let t_fun_to_code_eager_r tc =
   let env = FApi.tc1_env tc in
   let eg = tc1_as_eagerF tc in
+  let ml, mr = eg.eg_ml, eg.eg_mr in
   let (fl,fr) = eg.eg_fl, eg.eg_fr in
-  let ml, sl, rl, al = ToCodeLow.to_code env fl mleft in
-  let mr, sr, rr, ar = ToCodeLow.to_code env fr mright in
+  let (ml0, mlt), sl, rl, al = ToCodeLow.to_code env fl ml in
+  assert (EcIdent.id_equal ml0 ml);
+  let (mr0, mrt), sr, rr, ar = ToCodeLow.to_code env fr mr in
+  assert (EcIdent.id_equal mr0 mr);
   let spr =
     let s = PVM.empty in
-    let s = ToCodeLow.add_var_tuple env pv_arg mleft  al ml s in
-    let s = ToCodeLow.add_var_tuple env pv_arg mright ar mr s in
+    let s = ToCodeLow.add_var_tuple env pv_arg ml0  al ml s in
+    let s = ToCodeLow.add_var_tuple env pv_arg mr0 ar mr s in
     s in
   let spo =
     let s = PVM.empty in
-    let s = ToCodeLow.add_var env pv_res mleft  rl ml s in
-    let s = ToCodeLow.add_var env pv_res mright rr mr s in
+    let s = ToCodeLow.add_var env pv_res ml0  rl ml s in
+    let s = ToCodeLow.add_var env pv_res mr0 rr mr s in
     s in
-  let pre   = PVM.subst env spr eg.eg_pr in
-  let post  = PVM.subst env spo eg.eg_po in
+  let pre   = PVM.subst env spr (eg_pr eg).inv in
+  let post  = PVM.subst env spo (eg_po eg).inv in
   let concl =
-    f_equivS ml mr pre (s_seq eg.eg_sl sl) (s_seq sr eg.eg_sr) post in
+    f_equivS mlt mrt {ml;mr;inv=pre} (s_seq eg.eg_sl sl) (s_seq sr eg.eg_sr) {ml;mr;inv=post} in
   FApi.xmutate1 tc `FunToCode [concl]
 
 (* -------------------------------------------------------------------- *)
@@ -522,8 +532,11 @@ let t_fun_to_code_r tc =
 let t_fun_to_code = FApi.t_low0 "fun-to-code" t_fun_to_code_r
 
 (* -------------------------------------------------------------------- *)
-let t_fun_r inv tc =
+let t_fun_r (inv: inv) tc =
   let th tc =
+    let inv = match inv with
+      | Inv_ss inv -> inv
+      | Inv_ts _ -> tc_error !!tc "expected a single sided invariant" in
     let env = FApi.tc1_env tc in
     let h   = destr_hoareF (FApi.tc1_goal tc) in
       if   NormMp.is_abstract_fun h.hf_f env
@@ -531,6 +544,9 @@ let t_fun_r inv tc =
       else t_hoareF_fun_def tc
 
   and teh tc =
+    let inv = match inv with
+      | Inv_ss inv -> inv
+      | Inv_ts _ -> tc_error !!tc "expected a single sided invariant" in
     let env = FApi.tc1_env tc in
     let h   = destr_eHoareF (FApi.tc1_goal tc) in
       if   NormMp.is_abstract_fun h.ehf_f env
@@ -538,6 +554,9 @@ let t_fun_r inv tc =
       else t_ehoareF_fun_def tc
 
   and tbh tc =
+    let inv = match inv with
+      | Inv_ss inv -> inv
+      | Inv_ts _ -> tc_error !!tc "expected a single sided invariant" in
     let env = FApi.tc1_env tc in
     let h   = destr_bdHoareF (FApi.tc1_goal tc) in
       if   NormMp.is_abstract_fun h.bhf_f env
@@ -545,6 +564,9 @@ let t_fun_r inv tc =
       else t_bdhoareF_fun_def tc
 
   and te tc =
+    let inv = match inv with
+      | Inv_ts inv -> inv
+      | Inv_ss _ -> tc_error !!tc "expected a two sided invariant" in
     let env = FApi.tc1_env tc in
     let e   = destr_equivF (FApi.tc1_goal tc) in
       if   NormMp.is_abstract_fun e.ef_fl env
@@ -574,14 +596,16 @@ let process_fun_to_code tc =
 (* -------------------------------------------------------------------- *)
 let process_fun_upto_info (bad, p, q) tc =
   let hyps = FApi.tc1_hyps tc in
-  let env' = LDecl.inv_memenv hyps in
+  let ml, mr = EcIdent.create "&1", EcIdent.create "&2" in
+  let env' = LDecl.inv_memenv ml mr hyps in
   let p    = TTC.pf_process_form !!tc env' tbool p in
   let q    = q |> omap (TTC.pf_process_form !!tc env' tbool) |> odfl f_true in
+  let m = EcIdent.create "&hr" in
   let bad  =
-    let env' = LDecl.push_active (EcMemory.abstract EcFol.mhr) hyps in
+    let env' = LDecl.push_active_ss (EcMemory.abstract m) hyps in
     TTC.pf_process_form !!tc env' tbool bad
   in
-    (bad, p, q)
+    ({inv=bad;m}, {inv=p;ml;mr}, {inv=q;ml;mr})
 
 (* -------------------------------------------------------------------- *)
 let process_fun_upto info g =
@@ -592,27 +616,31 @@ let process_fun_upto info g =
 let process_fun_abs inv tc =
   let t_hoare tc =
     let hyps = FApi.tc1_hyps tc in
-    let env' = LDecl.inv_memenv1 hyps in
+    let m = EcIdent.create "&hr" in
+    let env' = LDecl.inv_memenv1 m hyps in
     let inv  = TTC.pf_process_form !!tc env' tbool inv in
-    t_hoareF_abs inv tc
+    t_hoareF_abs {inv;m} tc
 
   and t_ehoare tc =
     let hyps = FApi.tc1_hyps tc in
-    let env' = LDecl.inv_memenv1 hyps in
+    let m = EcIdent.create "&hr" in
+    let env' = LDecl.inv_memenv1 m hyps in
     let inv  = TTC.pf_process_xreal !!tc env' inv in
-    t_ehoareF_abs inv tc
+    t_ehoareF_abs {inv;m} tc
 
   and t_bdhoare tc =
     let hyps = FApi.tc1_hyps tc in
-    let env' = LDecl.inv_memenv1 hyps in
+    let m = EcIdent.create "&hr" in
+    let env' = LDecl.inv_memenv1 m hyps in
     let inv  = TTC.pf_process_form !!tc env' tbool inv in
-    t_bdhoareF_abs inv tc
+    t_bdhoareF_abs {inv;m} tc
 
   and t_equiv tc =
     let hyps = FApi.tc1_hyps tc in
-    let env' = LDecl.inv_memenv hyps in
+    let ml, mr = EcIdent.create "&1", EcIdent.create "&2" in
+    let env' = LDecl.inv_memenv ml mr hyps in
     let inv  = TTC.pf_process_form !!tc env' tbool inv in
-    t_equivF_abs inv tc
+    t_equivF_abs {inv;ml;mr} tc
 
   in
   t_hF_or_bhF_or_eF ~th:t_hoare ~teh:t_ehoare ~tbh:t_bdhoare ~te:t_equiv tc
