@@ -2,34 +2,25 @@
 open EcFol
 open EcCoreGoal
 open EcLowPhlGoal
-
-(*-------------------------------------------------------------------- *)
-let build_sym ml mr pr po =
-  let s = Fsubst.f_subst_id in
-  let s = Fsubst.f_bind_mem s ml mr in
-  let s = Fsubst.f_bind_mem s mr ml in
-  let s = Fsubst.f_subst s in
-  (s pr, s po)
+open EcAst
+open EcSubst
 
 (*-------------------------------------------------------------------- *)
 let t_equivF_sym tc =
-  let ef    = tc1_as_equivF tc in
-  let pr,po = build_sym mleft mright ef.ef_pr ef.ef_po in
-  let cond  = f_equivF pr ef.ef_fr ef.ef_fl po in
+  let ef     = tc1_as_equivF tc in
+  let ml, mr = ef.ef_ml, ef.ef_mr in
+  let pr     = {ml;mr;inv=(ts_inv_rebind (ef_pr ef) mr ml).inv} in
+  let po     = {ml;mr;inv=(ts_inv_rebind (ef_po ef) mr ml).inv} in
+  let cond   = f_equivF pr ef.ef_fr ef.ef_fl po in
   FApi.xmutate1 tc `EquivSym [cond]
 
 (*-------------------------------------------------------------------- *)
 let t_equivS_sym tc =
   let es    = tc1_as_equivS tc in
-  let pr,po = build_sym (fst es.es_ml) (fst es.es_mr) es.es_pr es.es_po in
-  let cond  = f_equivS_r {
-    es_ml = fst es.es_ml, snd es.es_mr;
-    es_mr = fst es.es_mr, snd es.es_ml;
-    es_sl = es.es_sr;
-    es_sr = es.es_sl;
-    es_pr = pr;
-    es_po = po; } in
-
+  let (ml, mtl), (mr, mtr) = es.es_ml, es.es_mr in
+  let pr    = {ml;mr;inv=(ts_inv_rebind (es_pr es) mr ml).inv} in
+  let po    = {ml;mr;inv=(ts_inv_rebind (es_po es) mr ml).inv} in
+  let cond  = f_equivS mtl mtr pr es.es_sr es.es_sl po in
   FApi.xmutate1 tc `EquivSym [cond]
 
 (*-------------------------------------------------------------------- *)
