@@ -22,14 +22,14 @@ open EcCoreGoal.FApi
 (* -------------------------------------------------------------------- *)
 val t_equivS_trans :
      EcMemory.memtype * EcModules.stmt
-  -> EcFol.form * EcFol.form
-  -> EcFol.form * EcFol.form
+  -> EcAst.ts_inv * EcAst.ts_inv
+  -> EcAst.ts_inv * EcAst.ts_inv
   -> EcCoreGoal.FApi.backward
 
 val t_equivF_trans :
      EcPath.xpath
-  -> EcFol.form * EcFol.form
-  -> EcFol.form * EcFol.form
+  -> EcAst.ts_inv * EcAst.ts_inv
+  -> EcAst.ts_inv * EcAst.ts_inv
   -> EcCoreGoal.FApi.backward
 
 (*---------------------------------------------------------------------------------------*)
